# Copyright (C) 2010,2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2018  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Main class for guest installation
"""

import base64
import errno
import hashlib
import logging
import os
import re
import shutil
import socket
import stat
import struct
import subprocess
import tempfile
import time
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse
import uuid

import M2Crypto

import guestfs

import libvirt

import lxml.etree

import oz.GuestFSManager
import oz.OzException
import oz.ozutil


class Guest(object):
    """
    Main class for guest installation.
    """
    def _discover_libvirt_type(self):
        """
        Internal method to discover the libvirt type (qemu, kvm, etc) that
        we should use, if not specified by the user.
        """
        if self.libvirt_type is None:
            doc = lxml.etree.fromstring(self.libvirt_conn.getCapabilities())

            # Libvirt calls the old intel 32-bit architecture i686, while we
            # refer to it as i386.  Do the mapping here, since we need to look
            # up the libvirt name.
            libvirtarch = self.tdl.arch
            if libvirtarch == 'i386':
                libvirtarch = 'i686'

            if len(doc.xpath("/capabilities/guest/arch[@name='%s']/domain[@type='kvm']" % (libvirtarch))) > 0:
                self.libvirt_type = 'kvm'
            elif len(doc.xpath("/capabilities/guest/arch[@name='%s']/domain[@type='qemu']" % (libvirtarch))) > 0:
                self.libvirt_type = 'qemu'
            else:
                raise oz.OzException.OzException("This host does not support virtualization type kvm or qemu for TDL arch (%s)" % (libvirtarch))

        self.log.debug("Libvirt type is %s", self.libvirt_type)

    def _discover_libvirt_bridge(self):
        """
        Internal method to discover a libvirt bridge (if necessary).
        """
        if self.bridge_name is None:
            # otherwise, try to detect a private libvirt bridge
            for netname in self.libvirt_conn.listNetworks():
                network = self.libvirt_conn.networkLookupByName(netname)

                xml = network.XMLDesc(0)
                doc = lxml.etree.fromstring(xml)

                forward = doc.xpath('/network/forward')
                if len(forward) != 1:
                    self.log.warning("Libvirt network without a forward element, skipping")
                    continue

                if forward[0].get('mode') == 'nat':
                    ips = doc.xpath('/network/ip')
                    if not ips:
                        self.log.warning("Libvirt network without an IP, skipping")
                        continue
                    for ip in ips:
                        family = ip.get("family")
                        if family is None or family == "ipv4":
                            self.bridge_name = network.bridgeName()
                            break

        if self.bridge_name is None:
            raise oz.OzException.OzException("Could not find a libvirt bridge.  Please run 'virsh net-start default' to start the default libvirt network, or see http://github.com/clalancette/oz/wiki/Oz-Network-Configuration for more information")

        self.log.debug("libvirt bridge name is %s", self.bridge_name)

    def connect_to_libvirt(self):
        """
        Method to connect to libvirt and detect various things about the
        environment.
        """
        def _libvirt_error_handler(ctxt, err):
            """
            Error callback to suppress libvirt printing to stderr by default.
            """
            pass

        libvirt.registerErrorHandler(_libvirt_error_handler, 'context')
        self.libvirt_conn = libvirt.open(self.libvirt_uri)
        self._discover_libvirt_bridge()
        self._discover_libvirt_type()

    def __init__(self, tdl, config, auto, output_disk, nicmodel, clockoffset,
                 mousetype, diskbus, iso_allowed, url_allowed, macaddress):
        self.tdl = tdl

        # for backwards compatibility
        self.name = self.tdl.name

        if self.tdl.arch not in ["i386", "x86_64", "ppc64", "ppc64le", "aarch64", "armv7l", "s390x"]:
            raise oz.OzException.OzException("Unsupported guest arch " + self.tdl.arch)

        if os.uname()[4] in ["i386", "i586", "i686"] and self.tdl.arch == "x86_64":
            raise oz.OzException.OzException("Host machine is i386, but trying to install x86_64 guest; this cannot work")

        self.log = logging.getLogger('%s.%s' % (__name__,
                                                self.__class__.__name__))
        self.uuid = uuid.uuid4()
        self.macaddr = macaddress
        if macaddress is None:
            self.macaddr = oz.ozutil.generate_macaddress()

        # configuration from 'paths' section
        self.output_dir = oz.ozutil.config_get_path(config, 'paths',
                                                    'output_dir',
                                                    oz.ozutil.default_output_dir())

        oz.ozutil.mkdir_p(self.output_dir)

        self.data_dir = oz.ozutil.config_get_path(config, 'paths',
                                                  'data_dir',
                                                  oz.ozutil.default_data_dir())

        self.screenshot_dir = oz.ozutil.config_get_path(config, 'paths',
                                                        'screenshot_dir',
                                                        oz.ozutil.default_screenshot_dir())

        self.sshprivkey = oz.ozutil.config_get_path(config, 'paths',
                                                    'sshprivkey',
                                                    oz.ozutil.default_sshprivkey())

        # configuration from 'libvirt' section
        self.libvirt_uri = oz.ozutil.config_get_key(config, 'libvirt', 'uri',
                                                    'qemu:///system')
        self.libvirt_type = oz.ozutil.config_get_key(config, 'libvirt', 'type',
                                                     None)
        self.bridge_name = oz.ozutil.config_get_key(config, 'libvirt',
                                                    'bridge_name', None)
        self.install_cpus = oz.ozutil.config_get_key(config, 'libvirt', 'cpus',
                                                     1)
        # the memory in the configuration file is specified in megabytes, but
        # libvirt expects kilobytes, so multiply by 1024
        if self.tdl.arch in ["ppc64", "ppc64le"]:
            # ppc64 needs at least 2Gb RAM
            self.install_memory = int(oz.ozutil.config_get_key(config, 'libvirt',
                                                               'memory', 2048)) * 1024
        else:
            self.install_memory = int(oz.ozutil.config_get_key(config, 'libvirt',
                                                               'memory', 1024)) * 1024
        self.image_type = oz.ozutil.config_get_key(config, 'libvirt',
                                                   'image_type', 'raw')

        # configuration from 'cache' section
        self.cache_original_media = oz.ozutil.config_get_boolean_key(config,
                                                                     'cache',
                                                                     'original_media',
                                                                     True)
        self.cache_modified_media = oz.ozutil.config_get_boolean_key(config,
                                                                     'cache',
                                                                     'modified_media',
                                                                     False)
        self.cache_jeos = oz.ozutil.config_get_boolean_key(config, 'cache',
                                                           'jeos', False)

        self.jeos_cache_dir = os.path.join(self.data_dir, "jeos")

        # configuration of "safe" ICICLE generation option
        self.safe_icicle_gen = oz.ozutil.config_get_boolean_key(config,
                                                                'icicle',
                                                                'safe_generation',
                                                                False)

        # configuration of 'timeouts' section
        self.default_install_timeout = int(oz.ozutil.config_get_key(config, 'timeouts',
                                                                    'install', 1200))
        self.inactivity_timeout = int(oz.ozutil.config_get_key(config, 'timeouts',
                                                               'inactivity', 300))
        self.boot_timeout = int(oz.ozutil.config_get_key(config, 'timeouts',
                                                         'boot', 300))
        self.shutdown_timeout = int(oz.ozutil.config_get_key(config, 'timeouts',
                                                             'shutdown', 90))

        # only pull a cached JEOS if it was built with the correct image type
        jeos_extension = self.image_type
        if self.image_type == 'raw':
            # backwards compatible
            jeos_extension = 'dsk'

        self.jeos_filename = os.path.join(self.jeos_cache_dir,
                                          self.tdl.distro + self.tdl.update + self.tdl.arch + '.' + jeos_extension)

        self.diskimage = output_disk
        if self.diskimage is None:
            ext = "." + self.image_type
            # compatibility with older versions of Oz
            if self.image_type == 'raw':
                ext = '.dsk'
            self.diskimage = os.path.join(self.output_dir, self.tdl.name + ext)

        if not os.path.isabs(self.diskimage):
            raise oz.OzException.OzException("Output disk image must be an absolute path")

        self.icicle_tmp = os.path.join(self.data_dir, "icicletmp",
                                       self.tdl.name)

        self.listen_port = oz.ozutil.get_free_port()

        self.console_listen_port = oz.ozutil.get_free_port()

        self.connect_to_libvirt()

        self.nicmodel = nicmodel
        if self.nicmodel is None:
            self.nicmodel = "rtl8139"
        self.clockoffset = clockoffset
        if self.clockoffset is None:
            self.clockoffset = "utc"
        self.mousetype = mousetype
        if self.mousetype is None:
            self.mousetype = "ps2"
        if diskbus is None or diskbus == "ide":
            self.disk_bus = "ide"
            self.disk_dev = "hda"
        elif diskbus == "virtio":
            self.disk_bus = "virtio"
            self.disk_dev = "vda"
        else:
            raise oz.OzException.OzException("Unknown diskbus type " + diskbus)

        self.rootpw = self.tdl.rootpw
        if self.rootpw is None:
            self.rootpw = "ozrootpw"

        try:
            # PPC64 directory structure differs from x86; disabling ISO boots
            if self.tdl.arch not in ["ppc64", "ppc64le"]:
                self.url = self._check_url(iso=iso_allowed, url=url_allowed)
            else:
                self.url = self._check_url(iso=False, url=url_allowed)
        except:
            self.log.debug("Install URL validation failed:", exc_info=True)
            raise

        oz.ozutil.mkdir_p(self.icicle_tmp)

        self.disksize = 10
        if self.tdl.disksize is not None:
            self.disksize = int(self.tdl.disksize)

        self.auto = auto
        if self.auto is None:
            self.auto = self.get_auto_path()

        self.log.debug("Name: %s, UUID: %s", self.tdl.name, self.uuid)
        self.log.debug("MAC: %s, distro: %s", self.macaddr, self.tdl.distro)
        self.log.debug("update: %s, arch: %s, diskimage: %s", self.tdl.update, self.tdl.arch, self.diskimage)
        self.log.debug("nicmodel: %s, clockoffset: %s", self.nicmodel, self.clockoffset)
        self.log.debug("mousetype: %s, disk_bus: %s, disk_dev: %s", self.mousetype, self.disk_bus, self.disk_dev)
        self.log.debug("icicletmp: %s, listen_port: %d", self.icicle_tmp, self.listen_port)
        self.log.debug("console_listen_port: %s", self.console_listen_port)

    def image_name(self):
        """
        Name of the image being built.
        """
        return self.name

    def output_image_path(self):
        """
        Path to the created image file.
        """
        return self.diskimage

    def get_auto_path(self):
        """
        Base method used to generate the path to the automatic installation
        file (kickstart, preseed, winnt.sif, etc).  Some subclasses override
        override this method to provide support for additional aliases.
        """
        return oz.ozutil.generate_full_auto_path(self.tdl.distro + self.tdl.update + ".auto")

    def default_auto_file(self):
        """
        Method to determine if the auto file is the default one or
        user-provided.
        """
        return self.auto == self.get_auto_path()

    def cleanup_old_guest(self):
        """
        Method to completely clean up an old guest, including deleting the
        disk file.  Use with caution!
        """
        self.log.info("Cleaning up guest named %s", self.tdl.name)
        try:
            dom = self.libvirt_conn.lookupByName(self.tdl.name)
            try:
                dom.destroy()
            except libvirt.libvirtError:
                pass
            dom.undefine()
        except libvirt.libvirtError:
            pass

        try:
            os.unlink(self.diskimage)
        except OSError as err:
            if err.errno != errno.ENOENT:
                raise

    def check_for_guest_conflict(self):
        """
        Method to check if any of our future actions will conflict with an
        already existing guest.  In particular, if a guest with the same
        name, UUID, or diskimage already exists, this throws an exception.
        """
        self.log.info("Checking for guest conflicts with %s", self.tdl.name)

        try:
            self.libvirt_conn.lookupByName(self.tdl.name)
            raise oz.OzException.OzException("Domain with name %s already exists" % (self.tdl.name))
        except libvirt.libvirtError:
            pass

        try:
            self.libvirt_conn.lookupByUUID(str(self.uuid))
            raise oz.OzException.OzException("Domain with UUID %s already exists" % (self.uuid))
        except libvirt.libvirtError:
            pass

        if os.access(self.diskimage, os.F_OK):
            raise oz.OzException.OzException("Diskimage %s already exists" % (self.diskimage))

    # the next 4 methods are intended to be overridden by the individual
    # OS backends; raise an error if they are called but not implemented

    def generate_install_media(self, force_download=False,
                               customize_or_icicle=False):
        """
        Base method for generating the install media for operating system
        installation.  This is expected to be overridden by all subclasses.
        """
        raise oz.OzException.OzException("Install media for %s%s is not implemented, install cannot continue" % (self.tdl.distro, self.tdl.update))

    def customize(self, libvirt_xml):
        """
        Base method for customizing the operating system.  This is expected
        to be overridden by subclasses that support customization.
        """
        raise oz.OzException.OzException("Customization for %s%s is not implemented" % (self.tdl.distro, self.tdl.update))

    def generate_icicle(self, libvirt_xml):
        """
        Base method for generating the ICICLE manifest from the operating
        system.  This is expect to be overridden by subclasses that support
        ICICLE generation.
        """
        raise oz.OzException.OzException("ICICLE generation for %s%s is not implemented" % (self.tdl.distro, self.tdl.update))

    # this method is intended to be an optimization if the user wants to do
    # both customize and generate_icicle
    def customize_and_generate_icicle(self, libvirt_xml):
        """
        Base method for doing operating system customization and ICICLE
        generation.  This is an optimization over doing the two steps
        separately for those classes that support customization and ICICLE
        generation.
        """
        raise oz.OzException.OzException("Customization and ICICLE generate for %s%s is not implemented" % (self.tdl.distro, self.tdl.update))

    class _InstallDev(object):
        """
        Class to hold information about an installation device.
        """
        def __init__(self, devicetype, path, bus):
            self.devicetype = devicetype
            self.path = path
            self.bus = bus

    def _generate_serial_xml(self, devices):
        """
        Method to generate the serial portion of the libvirt XML.
        """
        serial = oz.ozutil.lxml_subelement(devices, "serial", None, {'type': 'tcp'})
        oz.ozutil.lxml_subelement(serial, "source", None,
                                  {'mode': 'bind', 'host': '127.0.0.1', 'service': str(self.listen_port)})
        oz.ozutil.lxml_subelement(serial, "protocol", None, {'type': 'raw'})
        target = oz.ozutil.lxml_subelement(serial, "target", None, {'port': '1'})
        if self.tdl.arch == 's390x':
            # use a different console, as sclp can be used as most once
            target.set('type', 'sclp-serial')
            oz.ozutil.lxml_subelement(target, "model", None, {'name': 'sclplmconsole'})

    def _generate_virtio_channel(self, devices, name):
        """
        Method to generate libvirt XML for a virtio channel.
        """
        virtio = oz.ozutil.lxml_subelement(devices, "channel", None, {'type': 'tcp'})
        oz.ozutil.lxml_subelement(virtio, "source", None,
                                  {'mode': 'bind', 'host': '127.0.0.1', 'service': str(self.console_listen_port)})
        oz.ozutil.lxml_subelement(virtio, "target", None, {"type": "virtio", "name": name})

    def _generate_xml(self, bootdev, installdev, kernel=None, initrd=None,
                      cmdline=None, virtio_channel_name=None):
        """
        Method to generate libvirt XML useful for installation.
        """
        self.log.info("Generate XML for guest %s with bootdev %s", self.tdl.name, bootdev)

        # top-level domain element
        domain = lxml.etree.Element("domain", type=self.libvirt_type)
        # name element
        oz.ozutil.lxml_subelement(domain, "name", self.tdl.name)
        # memory elements
        oz.ozutil.lxml_subelement(domain, "memory", str(self.install_memory))
        oz.ozutil.lxml_subelement(domain, "currentMemory", str(self.install_memory))
        # uuid
        oz.ozutil.lxml_subelement(domain, "uuid", str(self.uuid))
        # clock offset
        oz.ozutil.lxml_subelement(domain, "clock", None, {'offset': self.clockoffset})
        # vcpu
        oz.ozutil.lxml_subelement(domain, "vcpu", str(self.install_cpus))
        # features
        features = oz.ozutil.lxml_subelement(domain, "features")
        oz.ozutil.lxml_subelement(features, "acpi")
        oz.ozutil.lxml_subelement(features, "apic")
        oz.ozutil.lxml_subelement(features, "pae")
        # CPU
        if self.tdl.arch in ["aarch64", "armv7l"] and self.libvirt_type == "kvm":
            # Possibly related to BZ 1171501 - need host passthrough for aarch64 and arm with kvm
            cpu = oz.ozutil.lxml_subelement(domain, "cpu", None, {'mode': 'custom', 'match': 'exact'})
            oz.ozutil.lxml_subelement(cpu, "model", "host", {'fallback': 'allow'})
        # os
        osNode = oz.ozutil.lxml_subelement(domain, "os")
        mods = None
        if self.tdl.arch in ["aarch64", "armv7l"]:
            mods = {'arch': self.tdl.arch, 'machine': 'virt'}
        oz.ozutil.lxml_subelement(osNode, "type", "hvm", mods)
        if bootdev:
            oz.ozutil.lxml_subelement(osNode, "boot", None, {'dev': bootdev})
        if kernel:
            oz.ozutil.lxml_subelement(osNode, "kernel", kernel)
        if initrd:
            oz.ozutil.lxml_subelement(osNode, "initrd", initrd)
        if cmdline:
            if self.tdl.arch == "armv7l":
                cmdline += " console=ttyAMA0"
            oz.ozutil.lxml_subelement(osNode, "cmdline", cmdline)
        if self.tdl.arch == "aarch64":
            loader, nvram = oz.ozutil.find_uefi_firmware(self.tdl.arch)
            oz.ozutil.lxml_subelement(osNode, "loader", loader, {'readonly': 'yes', 'type': 'pflash'})
            oz.ozutil.lxml_subelement(osNode, "nvram", None, {'template': nvram})
        # poweroff, reboot, crash
        oz.ozutil.lxml_subelement(domain, "on_poweroff", "destroy")
        oz.ozutil.lxml_subelement(domain, "on_reboot", "destroy")
        oz.ozutil.lxml_subelement(domain, "on_crash", "destroy")
        # devices
        devices = oz.ozutil.lxml_subelement(domain, "devices")
        # graphics
        if self.tdl.arch not in ["aarch64", "armv7l", "s390x"]:
            # qemu for arm/aarch64/s390x does not support a graphical console - amazingly
            oz.ozutil.lxml_subelement(devices, "graphics", None, {'port': '-1', 'type': 'vnc'})
        # network
        interface = oz.ozutil.lxml_subelement(devices, "interface", None, {'type': 'bridge'})
        oz.ozutil.lxml_subelement(interface, "source", None, {'bridge': self.bridge_name})
        oz.ozutil.lxml_subelement(interface, "mac", None, {'address': self.macaddr})
        oz.ozutil.lxml_subelement(interface, "model", None, {'type': self.nicmodel})
        # input
        mousedict = {'bus': self.mousetype}
        if self.mousetype == "ps2":
            mousedict['type'] = 'mouse'
        elif self.mousetype == "usb":
            mousedict['type'] = 'tablet'
        oz.ozutil.lxml_subelement(devices, "input", None, mousedict)
        # serial console pseudo TTY
        console = oz.ozutil.lxml_subelement(devices, "serial", None, {'type': 'pty'})
        oz.ozutil.lxml_subelement(console, "target", None, {'port': '0'})
        # serial
        self._generate_serial_xml(devices)
        # virtio
        if virtio_channel_name is not None:
            self._generate_virtio_channel(devices, virtio_channel_name)
            self.has_consolelog = True
        else:
            self.has_consolelog = False
        # virtio-rng
        virtioRNG = oz.ozutil.lxml_subelement(devices, "rng", None, {'model': 'virtio'})
        oz.ozutil.lxml_subelement(virtioRNG, "rate", None, {'bytes': '1024', 'period': '1000'})
        oz.ozutil.lxml_subelement(virtioRNG, "backend", "/dev/random", {'model': 'random'})
        # boot disk
        bootDisk = oz.ozutil.lxml_subelement(devices, "disk", None, {'device': 'disk', 'type': 'file'})
        oz.ozutil.lxml_subelement(bootDisk, "target", None, {'dev': self.disk_dev, 'bus': self.disk_bus})
        oz.ozutil.lxml_subelement(bootDisk, "source", None, {'file': self.diskimage})
        oz.ozutil.lxml_subelement(bootDisk, "driver", None, {'name': 'qemu', 'type': self.image_type})
        # install disk (if any)
        if not installdev:
            installdev_list = []
        elif not isinstance(installdev, list):
            installdev_list = [installdev]
        else:
            installdev_list = installdev
        for idev in installdev_list:
            install = oz.ozutil.lxml_subelement(devices, "disk", None, {'type': 'file', 'device': idev.devicetype})
            oz.ozutil.lxml_subelement(install, "source", None, {'file': idev.path})
            oz.ozutil.lxml_subelement(install, "target", None, {'dev': idev.bus})

        xml = lxml.etree.tostring(domain, pretty_print=True, encoding="unicode")
        self.log.debug("Generated XML:\n%s", xml)

        return xml

    def _internal_generate_diskimage(self, size=10, force=False,
                                     create_partition=False,
                                     image_filename=None,
                                     backing_filename=None):
        """
        Internal method to generate a diskimage.
        Set image_filename to override the default selection of self.diskimage
        Set backing_filename to force diskimage to be a writeable qcow2 snapshot
        backed by "backing_filename" which can be either a raw image or a
        qcow2 image.
        """
        if not force and os.access(self.jeos_filename, os.F_OK):
            # if we found a cached JEOS, we don't need to do anything here;
            # we'll copy the JEOS itself later on
            return

        self.log.info("Generating %dGB diskimage for %s", size, self.tdl.name)

        diskimage = self.diskimage
        if image_filename:
            diskimage = image_filename

        directory = os.path.dirname(diskimage)
        filename = os.path.basename(diskimage)

        # create the pool XML
        pool = lxml.etree.Element("pool", type="dir")
        oz.ozutil.lxml_subelement(pool, "name", "oztempdir" + str(uuid.uuid4()))
        target = oz.ozutil.lxml_subelement(pool, "target")
        oz.ozutil.lxml_subelement(target, "path", directory)
        pool_xml = lxml.etree.tostring(pool, pretty_print=True, encoding="unicode")

        # create the volume XML
        vol = lxml.etree.Element("volume", type="file")
        oz.ozutil.lxml_subelement(vol, "name", filename)
        oz.ozutil.lxml_subelement(vol, "allocation", "0")
        target = oz.ozutil.lxml_subelement(vol, "target")
        imgtype = self.image_type
        if backing_filename:
            # Only qcow2 supports image creation using a backing file
            imgtype = "qcow2"
        oz.ozutil.lxml_subelement(target, "format", None, {"type": imgtype})

        # FIXME: this makes the permissions insecure, but is needed since
        # libvirt launches guests as qemu:qemu
        permissions = oz.ozutil.lxml_subelement(target, "permissions")
        oz.ozutil.lxml_subelement(permissions, "mode", "0666")

        capacity = size
        if backing_filename:
            # FIXME: Revisit as BZ 958510 evolves
            # At the moment libvirt forces us to specify a size rather than
            # assuming we want to inherit the size of our backing file.
            # It may be possible to avoid this inspection step if libvirt
            # allows creation without an explicit capacity element.
            qcow_size = oz.ozutil.check_qcow_size(backing_filename)
            if qcow_size:
                capacity = qcow_size / 1024 / 1024 / 1024
                backing_format = 'qcow2'
            else:
                capacity = os.path.getsize(backing_filename) / 1024 / 1024 / 1024
                backing_format = 'raw'
            backing = oz.ozutil.lxml_subelement(vol, "backingStore")
            oz.ozutil.lxml_subelement(backing, "path", backing_filename)
            oz.ozutil.lxml_subelement(backing, "format", None,
                                      {"type": backing_format})

        oz.ozutil.lxml_subelement(vol, "capacity", str(capacity), {'unit': 'G'})
        vol_xml = lxml.etree.tostring(vol, pretty_print=True, encoding="unicode")

        # sigh.  Yes, this is racy; if a pool is defined during this loop, we
        # might miss it.  I'm not quite sure how to do it better, and in any
        # case we don't expect that to happen often
        started = False
        found = False
        for poolname in self.libvirt_conn.listDefinedStoragePools() + self.libvirt_conn.listStoragePools():
            pool = self.libvirt_conn.storagePoolLookupByName(poolname)
            doc = lxml.etree.fromstring(pool.XMLDesc(0))
            res = doc.xpath('/pool/target/path')
            if len(res) != 1:
                continue
            if res[0].text == directory:
                # OK, this pool manages that directory; make sure it is running
                found = True
                if not pool.isActive():
                    pool.create(0)
                    started = True
                break

        if not found:
            pool = self.libvirt_conn.storagePoolCreateXML(pool_xml, 0)
            started = True

        def _vol_create_cb(args):
            """
            The callback used for waiting on volume creation to complete.
            """
            pool = args[0]
            vol_xml = args[1]

            try:
                pool.refresh(0)
            except libvirt.libvirtError as e:
                if e.get_error_code() == libvirt.VIR_ERR_INTERNAL_ERROR:
                    # libvirt returns a VIR_ERR_INTERNAL_ERROR when the
                    # refresh fails.
                    return False
                raise

            # this is a bit complicated, because of the cases that can
            # happen.  The cases are:
            #
            # 1.  The volume did not exist.  In this case, storageVolLookupByName()
            #     throws an exception, which we just ignore.  We then go on to
            #     create the volume
            # 2.  The volume did exist.  In this case, storageVolLookupByName()
            #     returns a valid volume object, and then we delete the volume
            try:
                vol = pool.storageVolLookupByName(filename)
                vol.delete(0)
            except libvirt.libvirtError as e:
                if e.get_error_code() != libvirt.VIR_ERR_NO_STORAGE_VOL:
                    raise

            pool.createXML(vol_xml, 0)

            return True

        try:
            # libvirt will not allow us to do certain operations (like a refresh)
            # while other operations are happening on a pool (like creating a new
            # volume).  Since we don't exactly know which other processes might be
            # running on the system, we retry for a while until our refresh and
            # volume creation succeed.  In most cases these operations will be fast.
            oz.ozutil.timed_loop(90, _vol_create_cb, "Waiting for volume to be created", (pool, vol_xml))
        finally:
            if started:
                pool.destroy()

        if create_partition:
            if backing_filename:
                self.log.warning("Asked to create partition against a copy-on-write snapshot - ignoring")
            else:
                g_handle = oz.GuestFSManager.GuestFS(self.diskimage, self.image_type)
                g_handle.create_msdos_partition_table()
                g_handle.cleanup()

    def generate_diskimage(self, size=10, force=False):
        """
        Method to generate a diskimage.  By default, a blank diskimage of
        10GB will be created; the caller can override this with the size
        parameter, specified in GB.  If force is False (the default), then
        a diskimage will not be created if a cached JEOS is found.  If
        force is True, a diskimage will be created regardless of whether a
        cached JEOS exists.  See the oz-install man page for more
        information about JEOS caching.
        """
        return self._internal_generate_diskimage(size, force, False)

    def _get_disks_and_interfaces(self, libvirt_xml):
        """
        Method to figure out the disks and interfaces attached to a domain.
        The method returns two lists: the first is a list of disk devices (like
        hda, hdb, etc), and the second is a list of network devices (like vnet0,
        vnet1, etc).
        """
        doc = lxml.etree.fromstring(libvirt_xml)
        disktargets = doc.xpath("/domain/devices/disk/target")

        if len(disktargets) < 1:
            raise oz.OzException.OzException("Could not find disk target")
        disks = []
        for target in disktargets:
            dev = target.get('dev')
            if dev:
                disks.append(dev)
        if not disks:
            raise oz.OzException.OzException("Could not find disk target device")
        inttargets = doc.xpath("/domain/devices/interface/target")
        if len(inttargets) < 1:
            raise oz.OzException.OzException("Could not find interface target")
        interfaces = []
        for target in inttargets:
            dev = target.get('dev')
            if dev:
                interfaces.append(dev)
        if not interfaces:
            raise oz.OzException.OzException("Could not find interface target device")

        return disks, interfaces

    def _get_disk_and_net_activity(self, libvirt_dom, disks, interfaces):
        """
        Method to collect the disk and network activity by the domain.  The
        method returns two numbers: the first is the sum of all disk activity
        from all disks, and the second is the sum of all network traffic from
        all network devices.
        """
        total_disk_req = 0
        for dev in disks:
            rd_req, rd_bytes, wr_req, wr_bytes, errs = libvirt_dom.blockStats(dev)
            total_disk_req += rd_req + wr_req

        total_net_bytes = 0
        for dev in interfaces:
            rx_bytes, rx_packets, rx_errs, rx_drop, tx_bytes, tx_packets, tx_errs, tx_drop = libvirt_dom.interfaceStats(dev)
            total_net_bytes += rx_bytes + tx_bytes

        return total_disk_req, total_net_bytes

    def _wait_for_install_finish(self, xml, max_time):
        """
        Method to wait for an installation to finish.  This will wait around
        until either the VM has gone away (at which point it is assumed the
        install was successful), or until the timeout is reached (at which
        point it is assumed the install failed and raise an exception).
        """

        libvirt_dom = self.libvirt_conn.createXML(xml, 0)

        disks, interfaces = self._get_disks_and_interfaces(libvirt_dom.XMLDesc(0))

        self.last_disk_activity = 0
        self.last_network_activity = 0
        self.inactivity_countdown = self.inactivity_timeout
        self.saved_exception = None
        if self.has_consolelog:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.settimeout(1)
            self.sock.connect(('127.0.0.1', self.console_listen_port))

        def _finish_cb(self):
            '''
            A callback for _looper to deal with waiting for a guest to finish installing.
            '''
            if self.inactivity_countdown <= 0:
                return True
            try:
                total_disk_req, total_net_bytes = self._get_disk_and_net_activity(libvirt_dom, disks, interfaces)
            except libvirt.libvirtError as e:
                # we save the exception here because we want to raise it later
                # if this was a "real" exception
                self.saved_exception = e
                return True

            # rd_req and wr_req are the *total* number of disk read requests and
            # write requests ever made for this domain.  Similarly rd_bytes and
            # wr_bytes are the total number of network bytes read or written
            # for this domain

            # we define activity as having done a read or write request on the
            # install disk, or having done at least 4KB of network transfers in
            # the last second.  The thinking is that if the installer is putting
            # bits on disk, there will be disk activity, so we should keep
            # waiting.  On the other hand, the installer might be downloading
            # bits to eventually install on disk, so we look for network
            # activity as well.  We say that transfers of at least 4KB must be
            # made, however, to try to reduce false positives from things like
            # ARP requests

            if (total_disk_req == self.last_disk_activity) and (total_net_bytes < (self.last_network_activity + 4096)):
                # if we saw no read or write requests since the last iteration,
                # decrement our activity timer
                self.inactivity_countdown -= 1
            else:
                # if we did see some activity, then we can reset the timer
                self.inactivity_countdown = self.inactivity_timeout

            self.last_disk_activity = total_disk_req
            self.last_network_activity = total_net_bytes

            if self.has_consolelog:
                try:
                    # note that we have to build the data up here, since there
                    # is no guarantee that we will get the whole write in one go
                    data = self.sock.recv(65536)
                    if data:
                        self.log.debug(data.decode('utf-8'))
                except socket.timeout:
                    # the socket times out after 1 second.  We can just fall
                    # through to the below code because it is a noop.
                    pass

            return False

        finished = oz.ozutil.timed_loop(max_time, _finish_cb, "Waiting for %s to finish installing" % (self.tdl.name), self)

        # We get here because of a libvirt exception, an absolute timeout, or
        # an I/O timeout; we sort this out below
        if not finished:
            # if we timed out, then let's make sure to take a screenshot.
            screenshot_text = self._capture_screenshot(libvirt_dom)
            raise oz.OzException.OzException("Timed out waiting for install to finish.  %s" % (screenshot_text))
        elif self.inactivity_countdown <= 0:
            # if we saw no disk or network activity in the countdown window,
            # we presume the install has hung.  Fail here
            screenshot_text = self._capture_screenshot(libvirt_dom)
            raise oz.OzException.OzException("No disk activity in %d seconds, failing.  %s" % (self.inactivity_timeout, screenshot_text))

        # We get here only if we got a libvirt exception
        if not self._wait_for_guest_shutdown(libvirt_dom):
            if self.saved_exception is not None:
                self.log.debug("Libvirt Domain Info Failed:")
                self.log.debug(" code is %d", self.saved_exception.get_error_code())
                self.log.debug(" domain is %d", self.saved_exception.get_error_domain())
                self.log.debug(" message is %s", self.saved_exception.get_error_message())
                self.log.debug(" level is %d", self.saved_exception.get_error_level())
                self.log.debug(" str1 is %s", self.saved_exception.get_str1())
                self.log.debug(" str2 is %s", self.saved_exception.get_str2())
                self.log.debug(" str3 is %s", self.saved_exception.get_str3())
                self.log.debug(" int1 is %d", self.saved_exception.get_int1())
                self.log.debug(" int2 is %d", self.saved_exception.get_int2())
                raise self.saved_exception
            else:
                # the passed in exception was None, just raise a generic error
                raise oz.OzException.OzException("Unknown libvirt error")

        self.log.info("Install of %s succeeded", self.tdl.name)

    def _wait_for_guest_shutdown(self, libvirt_dom):
        """
        Method to wait around for orderly shutdown of a running guest.  Returns
        True if the guest shutdown in the specified time, False otherwise.
        """

        def _shutdown_cb(self):
            '''
            The _looper callback to wait for the guest to shutdown.
            '''
            try:
                libvirt_dom.info()
            except libvirt.libvirtError as e:
                if e.get_error_code() == libvirt.VIR_ERR_NO_DOMAIN:
                    # Here, the guest really, truly went away cleanly.  Thus, we know this
                    # loop has successfully completed, and we can return True.
                    return True

                # In all other cases, we got some kind of exception we didn't understand.
                # Fall through to the False return, and if we don't eventually see the
                # VIR_ERR_NO_DOMAIN, we'll return False to the caller.

            return False

        return oz.ozutil.timed_loop(self.shutdown_timeout, _shutdown_cb, "Waiting for %s to finish shutdown" % (self.tdl.name), self)

    def _get_csums(self, original_url, outdir, outputfd):
        """
        Internal method to fetch the checksum file and compute the checksum
        on the downloaded data.
        """
        if self.tdl.iso_md5_url:
            url = self.tdl.iso_md5_url
            hashname = 'md5'
        elif self.tdl.iso_sha1_url:
            url = self.tdl.iso_sha1_url
            hashname = 'sha1'
        elif self.tdl.iso_sha256_url:
            url = self.tdl.iso_sha256_url
            hashname = 'sha256'
        else:
            return True

        originalname = os.path.basename(urlparse.urlparse(original_url)[2])

        csumname = os.path.join(outdir,
                                self.tdl.distro + self.tdl.update + self.tdl.arch + "-CHECKSUM")

        (csumfd, outdir) = oz.ozutil.open_locked_file(csumname)

        os.ftruncate(csumfd, 0)

        try:
            self.log.debug("Checksum requested, fetching %s file", hashname)
            oz.ozutil.http_download_file(url, csumfd, False, self.log)
        finally:
            os.close(csumfd)

        upstream_sum = getattr(oz.ozutil,
                               'get_' + hashname + 'sum_from_file')(csumname, originalname)

        os.unlink(csumname)

        if not upstream_sum:
            raise oz.OzException.OzException("Could not find checksum for original file " + originalname)

        self.log.debug("Calculating checksum of downloaded file")
        os.lseek(outputfd, 0, os.SEEK_SET)

        local_sum = getattr(hashlib, hashname)()

        buf = oz.ozutil.read_bytes_from_fd(outputfd, 4096)
        while buf != '':
            local_sum.update(buf)
            buf = oz.ozutil.read_bytes_from_fd(outputfd, 4096)

        return local_sum.hexdigest() == upstream_sum

    def _get_original_media(self, url, fd, outdir, force_download):
        """
        Method to fetch the original media from url.  If the media is already
        cached locally, the cached copy will be used instead.
        """
        self.log.info("Fetching the original media")

        info = oz.ozutil.http_get_header(url)

        if 'HTTP-Code' not in info or int(info['HTTP-Code']) >= 400 or 'Content-Length' not in info or int(info['Content-Length']) < 0:
            raise oz.OzException.OzException("Could not reach %s to fetch boot media: %r" % (url, info))

        content_length = int(info['Content-Length'])

        if content_length == 0:
            raise oz.OzException.OzException("Install media of 0 size detected, something is wrong")

        if not force_download:
            if content_length == os.fstat(fd)[stat.ST_SIZE]:
                if self._get_csums(url, outdir, fd):
                    self.log.info("Original install media available, using cached version")
                    return

                self.log.info("Original available, but checksum mis-match; re-downloading")

        # before fetching everything, make sure that we have enough
        # space on the filesystem to store the data we are about to download
        devdata = os.statvfs(outdir)
        if (devdata.f_bsize * devdata.f_bavail) < content_length:
            raise oz.OzException.OzException("Not enough room on %s for install media" % (outdir))

        # at this point we know we are going to download something.  Make
        # sure to truncate the file so no stale data is left on the end
        os.ftruncate(fd, 0)

        self.log.info("Fetching the original install media from %s", url)
        oz.ozutil.http_download_file(url, fd, True, self.log)

        filesize = os.fstat(fd)[stat.ST_SIZE]

        if filesize != content_length:
            # if the length we downloaded is not the same as what we
            # originally saw from the headers, something went wrong
            raise oz.OzException.OzException("Expected to download %d bytes, downloaded %d" % (content_length, filesize))

        if not self._get_csums(url, outdir, fd):
            raise oz.OzException.OzException("Checksum for downloaded file does not match!")

    def _capture_screenshot(self, libvirt_dom):
        """
        Method to capture a screenshot of the VM.
        """

        # First check to ensure that we've got a graphics device that would even let
        # us take a screenshot.
        domxml = libvirt_dom.XMLDesc(0)
        doc = lxml.etree.fromstring(domxml)
        graphics = doc.xpath("/domain/devices/graphics")
        if not graphics:
            return "No graphics device found, screenshot skipped"

        oz.ozutil.mkdir_p(self.screenshot_dir)
        # create a new stream
        st = libvirt_dom.connect().newStream(0)

        # start the screenshot
        mimetype = libvirt_dom.screenshot(st, 0, 0)

        if mimetype == "image/x-portable-pixmap":
            ext = ".ppm"
        elif mimetype == "image/png":
            ext = ".png"
        else:
            return "Unknown screenshot type, failed to take screenshot"

        try:
            screenshot = os.path.realpath(os.path.join(self.screenshot_dir,
                                                       self.tdl.name + "-" + str(time.time()) + ext))

            def sink(stream, buf, opaque):  # pylint: disable=unused-argument
                """
                Function that is called back from the libvirt stream.
                """
                # opaque is the open file object
                return oz.ozutil.write_bytes_to_fd(opaque, buf)

            fd = os.open(screenshot, os.O_RDWR | os.O_CREAT)
            try:
                st.recvAll(sink, fd)
            finally:
                os.close(fd)

            st.finish()
            text = "Check screenshot at %s for more detail" % (screenshot)
        except Exception:
            text = "Failed to take screenshot"

        return text

    def _modify_libvirt_xml_for_serial(self, libvirt_xml):
        """
        Internal method to take input libvirt XML (which may have been provided
        by the user) and add an appropriate serial section so that guest
        announcement works properly.
        """
        input_doc = lxml.etree.fromstring(libvirt_xml)
        serialNode = input_doc.xpath("/domain/devices/serial")

        # we first go looking through the existing <serial> elements (if any);
        # if any exist on port 1, we delete it from the working XML and re-add
        # it below
        for serial in serialNode:
            target = serial.xpath('target')
            if len(target) != 1:
                raise oz.OzException.OzException("libvirt XML has a serial port with %d target(s), it is invalid" % (len(target)))
            if target[0].get('port') == "1":
                serial.getparent().remove(serial)
                break

        # at this point, the XML should be clean of any serial port=1 entries
        # and we can add the one we want
        devices = input_doc.xpath("/domain/devices")
        devlen = len(devices)
        if devlen == 0:
            raise oz.OzException.OzException("No devices section specified, something is wrong with the libvirt XML")
        elif devlen > 1:
            raise oz.OzException.OzException("%d devices sections specified, something is wrong with the libvirt XML" % (devlen))

        self._generate_serial_xml(devices[0])

        xml = lxml.etree.tostring(input_doc, pretty_print=True, encoding="unicode")
        self.log.debug("Generated XML:\n%s", xml)
        return xml

    def _modify_libvirt_xml_diskimage(self, libvirt_xml, new_diskimage, image_type):
        """
        Internal method to take input libvirt XML and replace the existing disk
        image details with a new disk image file and, potentially, disk image
        type.  Used in safe ICICLE generation to replace the "real" disk image
        file with a temporary writeable snapshot.
        """
        self.log.debug("Modifying libvirt XML to use disk image (%s) of type (%s)", new_diskimage, image_type)
        input_doc = lxml.etree.fromstring(libvirt_xml)
        disks = input_doc.xpath("/domain/devices/disk")
        if len(disks) != 1:
            self.log.warning("Oz given a libvirt domain with more than 1 disk; using the first one parsed")

        source = disks[0].xpath('source')
        if len(source) != 1:
            raise oz.OzException.OzException("invalid <disk> entry without a source")
        source[0].set('file', new_diskimage)

        driver = disks[0].xpath('driver')
        # at the time this function was added, all boot disk device stanzas
        # have a driver section - even raw images
        if len(driver) == 1:
            driver[0].set('type', image_type)
        else:
            raise oz.OzException.OzException("Found a disk with an unexpected number of driver sections")

        xml = lxml.etree.tostring(input_doc, pretty_print=True, encoding="unicode")
        self.log.debug("Generated XML:\n%s", xml)
        return xml

    def _wait_for_guest_boot(self, libvirt_dom):
        """
        Method to wait around for a guest to boot.  Orderly guests will boot
        up and announce their presence via a TCP message; if that happens within
        the timeout, this method returns the IP address of the guest.  If that
        doesn't happen an exception is raised.
        """
        self.log.info("Waiting for guest %s to boot", self.tdl.name)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        try:
            self.sock.settimeout(1)
            self.sock.connect(('127.0.0.1', self.listen_port))

            self.addr = None
            self.data = b''

            def _boot_cb(self):
                '''
                The _looper callback to look for the guest to boot.
                '''
                try:
                    # note that we have to build the data up here, since there
                    # is no guarantee that we will get the whole write in one go
                    self.data += self.sock.recv(8192)
                except socket.timeout:
                    # the socket times out after 1 second.  We can just fall
                    # through to the below code because it is a noop.
                    pass

                # OK, we got data back from the socket.  Check to see if it is
                # is what we expect; essentially, some up-front garbage,
                # followed by a !<ip>,<uuid>!
                # Exclude ! from the wildcard to avoid errors when receiving two
                # announce messages in the same string
                match = re.search(b"!([^!]*?,[^!]*?)!$", self.data)
                if match is not None:
                    if len(match.groups()) != 1:
                        raise oz.OzException.OzException("Guest checked in with no data")
                    split = match.group(1).split(b',')
                    if len(split) != 2:
                        raise oz.OzException.OzException("Guest checked in with bogus data")
                    self.addr = split[0]
                    uuidstr = split[1]
                    try:
                        # use socket.inet_aton() to validate the IP address
                        socket.inet_aton(self.addr.decode('utf-8'))
                    except socket.error:
                        raise oz.OzException.OzException("Guest checked in with invalid IP address")

                    if uuidstr.decode('utf-8') != str(self.uuid):
                        raise oz.OzException.OzException("Guest checked in with unknown UUID")
                    return True

                # if the data we got didn't match, we need to continue waiting.
                # before going to sleep, make sure that the domain is still
                # around
                libvirt_dom.info()

                return False

            oz.ozutil.timed_loop(self.boot_timeout, _boot_cb, "Waiting for %s to finish boot" % (self.tdl.name), self)

        finally:
            self.sock.close()

        if self.addr is None:
            raise oz.OzException.OzException("Timed out waiting for guest to boot")

        self.addr = self.addr.decode('utf-8')
        self.log.debug("IP address of guest is %s", self.addr)

        return self.addr

    def _output_icicle_xml(self, lines, description, extra=None):
        """
        Generate ICICLE XML based on the data supplied.  The parameter 'lines'
        is expected to be an list of strings, with one package per list item.
        The parameter 'description' is a description of the guest.  The
        parameter 'extra' is an optional one that describes any additional
        information the user wanted in the ICICLE output.
        """
        icicle = lxml.etree.Element("icicle")
        if description is not None:
            oz.ozutil.lxml_subelement(icicle, "description", description)
        packages = oz.ozutil.lxml_subelement(icicle, "packages")

        for index, line in enumerate(lines):
            if line == "":
                continue
            package = oz.ozutil.lxml_subelement(packages, "package", None,
                                                {'name': line})
            if extra is not None:
                oz.ozutil.lxml_subelement(package, "extra", extra[index])

        return lxml.etree.tostring(icicle, pretty_print=True, encoding="unicode")

    def _check_url(self, iso=True, url=True):
        """
        Method to check that a TDL URL meets the requirements for a particular
        operating system.
        """

        # this method is *slightly* odd in that it references ISOs from the
        # generic Guest class.  However, the installtype comes from the user
        # TDL, which means that they could have specified an ISO installtype for
        # a floppy guest (for instance).  Since we internally always set
        # iso=False for floppy guests, this will raise an appropriate error

        if iso and self.tdl.installtype == 'iso':
            url = self.tdl.iso
        elif url and self.tdl.installtype == 'url':
            url = self.tdl.url

            # when doing URL installs, we can't allow localhost URLs (the URL
            # will be embedded into the installer, so the install is guaranteed
            # to fail with localhost URLs).  Disallow them here
            if urlparse.urlparse(url)[1] in ["localhost", "127.0.0.1",
                                             "localhost.localdomain"]:
                raise oz.OzException.OzException("Can not use localhost for an URL based install")
        else:
            if iso and url:
                raise oz.OzException.OzException("%s installs must be done via url or iso" % (self.tdl.distro))
            elif iso:
                raise oz.OzException.OzException("%s installs must be done via iso" % (self.tdl.distro))
            elif url:
                raise oz.OzException.OzException("%s installs for %s must be done via url" % (self.tdl.distro, self.tdl.arch))
            else:
                raise oz.OzException.OzException("Unknown error occurred while determining install URL")

        return url

    def _generate_openssh_key(self, privname):
        """
        Method to generate an OpenSSH compatible public/private keypair.
        """
        self.log.info("Generating new openssh key")
        pubname = privname + ".pub"
        if os.access(privname, os.F_OK) and not os.access(pubname, os.F_OK):
            # hm, private key exists but not public?  We have to regenerate
            os.unlink(privname)

        if not os.access(privname, os.F_OK) and os.access(pubname, os.F_OK):
            # hm, public key exists but not private?  We have to regenerate
            os.unlink(pubname)

        # when we get here, either both the private and public key exist, or
        # neither exist.  If they don't exist, generate them
        if not os.access(privname, os.F_OK) and not os.access(pubname, os.F_OK):
            def _null_callback(p, n, out):  # pylint: disable=unused-argument
                """
                Method to silence the default M2Crypto.RSA.gen_key output.
                """
                pass

            pubname = privname + '.pub'

            key = M2Crypto.RSA.gen_key(2048, 65537, _null_callback)

            # this is the binary public key, in ssh "BN" (BigNumber) MPI format.
            # The ssh BN MPI format consists of 4 bytes that describe the length
            # of the following data, followed by the data itself in big-endian
            # format.  The start of the string is 0x0007, which represent the 7
            # bytes following that make up 'ssh-rsa'.  The key exponent and
            # modulus as fetched out of M2Crypto are already in MPI format, so
            # we can just use them as-is.  We then have to base64 encode the
            # result, add a little header information, and then we have a
            # full public key.
            pubkey = b'\x00\x00\x00\x07' + b'ssh-rsa' + key.e + key.n

            username = os.getlogin()
            hostname = os.uname()[1]
            keystring = 'ssh-rsa %s %s@%s\n' % (base64.b64encode(pubkey).decode('utf-8'),
                                                username, hostname)

            oz.ozutil.mkdir_p(os.path.dirname(privname))
            key.save_key(privname, cipher=None)
            os.chmod(privname, 0o600)
            with open(pubname, 'w') as f:
                f.write(keystring)
            os.chmod(pubname, 0o644)


class CDGuest(Guest):
    """
    Class for guest installation via ISO.
    """
    class _PrimaryVolumeDescriptor(object):
        """
        Class to hold information about a CD's Primary Volume Descriptor.
        """
        def __init__(self, version, sysid, volid, space_size, set_size, seqnum):
            self.version = version
            self.system_identifier = sysid
            self.volume_identifier = volid
            self.space_size = space_size
            self.set_size = set_size
            self.seqnum = seqnum

    def __init__(self, tdl, config, auto, output_disk, nicmodel, clockoffset,
                 mousetype, diskbus, iso_allowed, url_allowed, macaddress):
        Guest.__init__(self, tdl, config, auto, output_disk, nicmodel,
                       clockoffset, mousetype, diskbus, iso_allowed,
                       url_allowed, macaddress)

        self.orig_iso = os.path.join(self.data_dir, "isos",
                                     self.tdl.distro + self.tdl.update + self.tdl.arch + "-" + self.tdl.installtype + ".iso")
        self.modified_iso_cache = os.path.join(self.data_dir, "isos",
                                               self.tdl.distro + self.tdl.update + self.tdl.arch + "-" + self.tdl.installtype + "-oz.iso")
        self.output_iso = os.path.join(self.output_dir,
                                       self.tdl.name + "-" + self.tdl.installtype + "-oz.iso")
        self.iso_contents = os.path.join(self.data_dir, "isocontent",
                                         self.tdl.name + "-" + self.tdl.installtype)
        # An optional floppy disk to be used for installation
        self.output_floppy = os.path.join(self.output_dir,
                                          self.tdl.name + "-" + self.tdl.installtype + "-oz.img")

        self.log.debug("Original ISO path: %s", self.orig_iso)
        self.log.debug("Modified ISO cache: %s", self.modified_iso_cache)
        self.log.debug("Output ISO path: %s", self.output_iso)
        self.log.debug("ISO content path: %s", self.iso_contents)

    def _get_original_iso(self, isourl, fd, outdir, force_download):
        """
        Method to fetch the original ISO for an operating system.
        """
        self._get_original_media(isourl, fd, outdir, force_download)

    def _copy_iso(self):
        """
        Method to copy the data out of an ISO onto the local filesystem.
        """
        self.log.info("Copying ISO contents for modification")
        try:
            shutil.rmtree(self.iso_contents)
        except OSError as err:
            if err.errno != errno.ENOENT:
                raise
        os.makedirs(self.iso_contents)

        self.log.info("Setting up guestfs handle for %s", self.tdl.name)
        gfs = guestfs.GuestFS()
        self.log.debug("Adding ISO image %s", self.orig_iso)
        gfs.add_drive_opts(self.orig_iso, readonly=1, format='raw')
        self.log.debug("Launching guestfs")
        gfs.launch()
        try:
            self.log.debug("Mounting ISO")
            gfs.mount_options('ro', "/dev/sda", "/")

            self.log.debug("Checking if there is enough space on the filesystem")
            isostat = gfs.statvfs("/")
            outputstat = os.statvfs(self.iso_contents)
            if (outputstat.f_bsize * outputstat.f_bavail) < (isostat['blocks'] * isostat['bsize']):
                raise oz.OzException.OzException("Not enough room on %s to extract install media" % (self.iso_contents))

            self.log.debug("Extracting ISO contents")
            current = os.getcwd()
            os.chdir(self.iso_contents)
            try:
                rd, wr = os.pipe()

                try:
                    # NOTE: it is very, very important that we use temporary
                    # files for collecting stdout and stderr here.  There is a
                    # nasty bug in python subprocess; if your process produces
                    # more than 64k of data on an fd that is using
                    # subprocess.PIPE, the whole thing will hang. To avoid
                    # this, we use temporary fds to capture the data
                    stdouttmp = tempfile.TemporaryFile()
                    stderrtmp = tempfile.TemporaryFile()

                    try:
                        tar = subprocess.Popen(["tar", "-x", "-v"], stdin=rd,
                                               stdout=stdouttmp,
                                               stderr=stderrtmp)
                        try:
                            gfs.tar_out("/", "/dev/fd/%d" % wr)
                        except:
                            # we need this here if gfs.tar_out throws an
                            # exception.  In that case, we need to manually
                            # kill off the tar process and re-raise the
                            # exception, otherwise we hang forever
                            tar.kill()
                            raise

                        # Make sure tar is fully done with the archive before
                        # we can continue.
                        tar.wait()
                        if tar.returncode:
                            self.log.debug('tar stdout: %s', stdouttmp.read())
                            self.log.debug('tar stderr: %s', stderrtmp.read())
                            raise oz.OzException.OzException("Tar exited with an error")

                    finally:
                        stdouttmp.close()
                        stderrtmp.close()
                finally:
                    os.close(rd)
                    os.close(wr)

                # since we extracted from an ISO, there are no write bits on
                # any of the directories.  Fix that here.
                oz.ozutil.recursively_add_write_bit(self.iso_contents)
            finally:
                os.chdir(current)
        finally:
            gfs.sync()
            gfs.umount_all()
            gfs.kill_subprocess()

    def _get_primary_volume_descriptor(self, cdfd):
        """
        Method to extract the primary volume descriptor from a CD.
        """
        # check out the primary volume descriptor to make sure it is sane
        cdfd.seek(16 * 2048)
        fmt = b"=B5sBB32s32sQLL32sHHHH"
        (desc_type, identifier, version, unused1, system_identifier,
         volume_identifier, unused2, space_size_le, space_size_be_unused,
         unused3, set_size_le, set_size_be, seqnum_le,
         seqnum_be) = struct.unpack(fmt, cdfd.read(struct.calcsize(fmt)))

        if desc_type != 0x1:
            raise oz.OzException.OzException("Invalid primary volume descriptor")
        if identifier != "CD001":
            raise oz.OzException.OzException("invalid CD isoIdentification")
        if unused1 != 0x0:
            raise oz.OzException.OzException("data in unused field")
        if unused2 != 0x0:
            raise oz.OzException.OzException("data in 2nd unused field")

        return self._PrimaryVolumeDescriptor(version, system_identifier,
                                             volume_identifier, space_size_le,
                                             set_size_le, seqnum_le)

    def _geteltorito(self, cdfile, outfile):
        """
        Method to extract the El-Torito boot sector off of a CD and write it
        to a file.
        """
        if cdfile is None:
            raise oz.OzException.OzException("input iso is None")
        if outfile is None:
            raise oz.OzException.OzException("output file is None")

        with open(cdfile, 'rb') as cdfd:
            self._get_primary_volume_descriptor(cdfd)

            # the 17th sector contains the boot specification and the offset of the
            # boot sector
            cdfd.seek(17 * 2048)

            # NOTE: With "native" alignment (the default for struct), there is
            # some padding that happens that causes the unpacking to fail.
            # Instead we force "standard" alignment, which has no padding
            fmt = "=B5sB23s41sI"
            (boot, isoIdent, version, toritoSpec, unused, bootP) = struct.unpack(fmt,
                                                                                 cdfd.read(struct.calcsize(fmt)))
            if boot != 0x0:
                raise oz.OzException.OzException("invalid CD boot sector")
            if isoIdent != "CD001":
                raise oz.OzException.OzException("invalid CD isoIdentification")
            if version != 0x1:
                raise oz.OzException.OzException("invalid CD version")
            if toritoSpec != "EL TORITO SPECIFICATION":
                raise oz.OzException.OzException("invalid CD torito specification")

            # OK, this looks like a bootable CD.  Seek to the boot sector, and
            # look for the header, 0x55, and 0xaa in the first 32 bytes
            cdfd.seek(bootP * 2048)
            fmt = "=BBH24sHBB"
            bootdata = cdfd.read(struct.calcsize(fmt))
            (header, platform, unused, manu_unused, unused2, five, aa) = struct.unpack(fmt,
                                                                                       bootdata)
            if header != 0x1:
                raise oz.OzException.OzException("invalid CD boot sector header")
            if platform != 0x0 and platform != 0x1 and platform != 0x2:
                raise oz.OzException.OzException("invalid CD boot sector platform")
            if unused != 0x0:
                raise oz.OzException.OzException("invalid CD unused boot sector field")
            if five != 0x55 or aa != 0xaa:
                raise oz.OzException.OzException("invalid CD boot sector footer")

            def _checksum(data):
                """
                Method to compute the checksum on the ISO.  Note that this is *not*
                a 1's complement checksum; when an addition overflows, the carry
                bit is discarded, not added to the end.
                """
                s = 0
                for i in range(0, len(data), 2):
                    w = ord(data[i]) + (ord(data[i + 1]) << 8)
                    s = (s + w) & 0xffff
                return s

            csum = _checksum(bootdata)
            if csum != 0:
                raise oz.OzException.OzException("invalid CD checksum: expected 0, saw %d" % (csum))

            # OK, everything so far has checked out.  Read the default/initial
            # boot entry
            cdfd.seek(bootP * 2048 + 32)
            fmt = "=BBHBBHIB"
            (boot, media, loadsegment_unused, systemtype_unused, unused, scount,
             imgstart, unused2) = struct.unpack(fmt, cdfd.read(struct.calcsize(fmt)))

            if boot != 0x88:
                raise oz.OzException.OzException("invalid CD initial boot indicator")
            if unused != 0x0 or unused2 != 0x0:
                raise oz.OzException.OzException("invalid CD initial boot unused field")

            if media == 0 or media == 4:
                count = scount
            elif media == 1:
                # 1.2MB floppy in sectors
                count = 1200 * 1024 / 512
            elif media == 2:
                # 1.44MB floppy in sectors
                count = 1440 * 1024 / 512
            elif media == 3:
                # 2.88MB floppy in sectors
                count = 2880 * 1024 / 512
            else:
                raise oz.OzException.OzException("invalid CD media type")

            # finally, seek to "imgstart", and read "count" sectors, which
            # contains the boot image
            cdfd.seek(imgstart * 2048)

            # The eltorito specification section 2.5 says:
            #
            # Sector Count. This is the number of virtual/emulated sectors the
            # system will store at Load Segment during the initial boot
            # procedure.
            #
            # and then Section 1.5 says:
            #
            # Virtual Disk - A series of sectors on the CD which INT 13 presents
            # to the system as a drive with 200 byte virtual sectors. There
            # are 4 virtual sectors found in each sector on a CD.
            #
            # (note that the bytes above are in hex).  So we read count*512
            eltoritodata = cdfd.read(count * 512)

        with open(outfile, "w") as f:
            f.write(eltoritodata)

    def _do_install(self, timeout=None, force=False, reboots=0,
                    kernelfname=None, ramdiskfname=None, cmdline=None,
                    extrainstalldevs=None, virtio_channel_name=None):
        """
        Internal method to actually run the installation.
        """
        if not force and os.access(self.jeos_filename, os.F_OK):
            self.log.info("Found cached JEOS (%s), using it", self.jeos_filename)
            oz.ozutil.copyfile_sparse(self.jeos_filename, self.diskimage)
            return self._generate_xml("hd", None)

        self.log.info("Running install for %s", self.tdl.name)

        if timeout is None:
            timeout = self.default_install_timeout

        cddev = self._InstallDev("cdrom", self.output_iso, "hdc")
        if extrainstalldevs is not None:
            extrainstalldevs.append(cddev)
            cddev = extrainstalldevs
        reboots_to_go = reboots
        while reboots_to_go >= 0:
            # if reboots_to_go is the same as reboots, it means that this is
            # the first time through and we should generate the "initial" xml
            if reboots_to_go == reboots:
                if kernelfname and os.access(kernelfname, os.F_OK) and ramdiskfname and os.access(ramdiskfname, os.F_OK) and cmdline:
                    xml = self._generate_xml(None, None, kernelfname,
                                             ramdiskfname, cmdline,
                                             virtio_channel_name)
                else:
                    xml = self._generate_xml("cdrom", cddev, None, None, None,
                                             virtio_channel_name)
            else:
                xml = self._generate_xml("hd", cddev, None, None, None,
                                         virtio_channel_name)

            self._wait_for_install_finish(xml, timeout)

            reboots_to_go -= 1

        if self.cache_jeos:
            self.log.info("Caching JEOS")
            oz.ozutil.mkdir_p(self.jeos_cache_dir)
            oz.ozutil.copyfile_sparse(self.diskimage, self.jeos_filename)

        return self._generate_xml("hd", None)

    def install(self, timeout=None, force=False):
        """
        Method to run the operating system installation.
        """
        return self._do_install(timeout, force, 0)

    def _check_pvd(self):
        """
        Base method to check the Primary Volume Descriptor on the ISO.  In the
        common case, do nothing; subclasses that need to check the media will
        override this.
        """
        pass

    def _check_iso_tree(self, customize_or_icicle):
        """
        Base method to check the exploded ISO tree.  In the common case, do
        nothing; subclasses that need to check the tree will override this.
        """
        pass

    def _add_iso_extras(self):
        """
        Method to modify the ISO based on the directories specified in the TDL
        file. This modification is done before the final OS and is not
        expected to be override by subclasses
        """
        for isoextra in self.tdl.isoextras:
            targetabspath = os.path.join(self.iso_contents,
                                         isoextra.destination)
            oz.ozutil.mkdir_p(os.path.dirname(targetabspath))

            parsedurl = urlparse.urlparse(isoextra.source)
            if parsedurl.scheme == 'file':
                if isoextra.element_type == "file":
                    oz.ozutil.copyfile_sparse(parsedurl.path, targetabspath)
                else:
                    oz.ozutil.copytree_merge(parsedurl.path, targetabspath)
            elif parsedurl.scheme == "ftp":
                if isoextra.element_type == "file":
                    fd = os.open(targetabspath,
                                 os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
                    try:
                        oz.ozutil.http_download_file(isoextra.source, fd, True,
                                                     self.log)
                    finally:
                        os.close(fd)
                else:
                    oz.ozutil.ftp_download_directory(parsedurl.hostname,
                                                     parsedurl.username,
                                                     parsedurl.password,
                                                     parsedurl.path,
                                                     targetabspath,
                                                     parsedurl.port)
            elif parsedurl.scheme == "http":
                if isoextra.element_type == "directory":
                    raise oz.OzException.OzException("ISO extra directories cannot be fetched over HTTP")
                else:
                    fd = os.open(targetabspath,
                                 os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
                    try:
                        oz.ozutil.http_download_file(isoextra.source, fd, True,
                                                     self.log)
                    finally:
                        os.close(fd)
            else:
                raise oz.OzException.OzException("The protocol '%s' is not supported for fetching remote files or directories" % parsedurl.scheme)

    def _modify_iso(self):
        """
        Base method to modify the ISO.  Subclasses are expected to override
        this.
        """
        raise oz.OzException.OzException("Internal error, subclass didn't override modify_iso")

    def _generate_new_iso(self):
        """
        Base method to generate the new ISO.  Subclasses are expected to
        override this.
        """
        raise oz.OzException.OzException("Internal error, subclass didn't override generate_new_iso")

    def _iso_generate_install_media(self, url, force_download,
                                    customize_or_icicle):
        """
        Method to generate the modified media necessary for unattended installs.
        """
        self.log.info("Generating install media")

        if not force_download:
            if os.access(self.jeos_filename, os.F_OK):
                # if we found a cached JEOS, we don't need to do anything here;
                # we'll copy the JEOS itself later on
                return
            elif os.access(self.modified_iso_cache, os.F_OK):
                self.log.info("Using cached modified media")
                shutil.copyfile(self.modified_iso_cache, self.output_iso)
                return

        (fd, outdir) = oz.ozutil.open_locked_file(self.orig_iso)

        try:
            self._get_original_iso(url, fd, outdir, force_download)
            self._check_pvd()
            self._copy_iso()

            # from here on out, we have to make sure to cleanup the exploded ISO
            try:
                self._check_iso_tree(customize_or_icicle)
                self._add_iso_extras()
                self._modify_iso()
                self._generate_new_iso()
                if self.cache_modified_media:
                    self.log.info("Caching modified media for future use")
                    shutil.copyfile(self.output_iso, self.modified_iso_cache)
            finally:
                self._cleanup_iso()
        finally:
            os.close(fd)

    def generate_install_media(self, force_download=False,
                               customize_or_icicle=False):
        """
        Method to generate the install media for the operating
        system.  If force_download is False (the default), then the
        original media will only be fetched if it is not cached locally.  If
        force_download is True, then the original media will be downloaded
        regardless of whether it is cached locally.
        """
        return self._iso_generate_install_media(self.url, force_download,
                                                customize_or_icicle)

    def _cleanup_iso(self):
        """
        Method to cleanup the local ISO contents.
        """
        self.log.info("Cleaning up old ISO data")
        # if we are running as non-root, then there might be some files left
        # around that are not writable, which means that the rmtree below would
        # fail.  Recurse into the iso_contents tree, doing a chmod +w on
        # every file and directory to make sure the rmtree succeeds
        oz.ozutil.recursively_add_write_bit(self.iso_contents)

        oz.ozutil.rmtree_and_sync(self.iso_contents)

    def cleanup_install(self):
        """
        Method to cleanup any transient install data.
        """
        self.log.info("Cleaning up after install")

        try:
            os.unlink(self.output_iso)
        except:
            pass

        try:
            os.unlink(self.output_floppy)
        except:
            pass

        if not self.cache_original_media:
            try:
                os.unlink(self.orig_iso)
            except:
                pass


class FDGuest(Guest):
    """
    Class for guest installation via floppy disk.
    """
    def __init__(self, tdl, config, auto, output_disk, nicmodel, clockoffset,
                 mousetype, diskbus, macaddress):
        Guest.__init__(self, tdl, config, auto, output_disk, nicmodel,
                       clockoffset, mousetype, diskbus, False, True, macaddress)
        self.orig_floppy = os.path.join(self.data_dir, "floppies",
                                        self.tdl.distro + self.tdl.update + self.tdl.arch + ".img")
        self.modified_floppy_cache = os.path.join(self.data_dir, "floppies",
                                                  self.tdl.distro + self.tdl.update + self.tdl.arch + "-oz.img")
        self.output_floppy = os.path.join(self.output_dir,
                                          self.tdl.name + "-oz.img")
        self.floppy_contents = os.path.join(self.data_dir, "floppycontent",
                                            self.tdl.name)

        self.log.debug("Original floppy path: %s", self.orig_floppy)
        self.log.debug("Modified floppy cache: %s", self.modified_floppy_cache)
        self.log.debug("Output floppy path: %s", self.output_floppy)
        self.log.debug("Floppy content path: %s", self.floppy_contents)

    def _get_original_floppy(self, floppyurl, fd, outdir, force_download):
        """
        Method to download the original floppy if necessary.
        """
        self._get_original_media(floppyurl, fd, outdir, force_download)

    def _copy_floppy(self):
        """
        Method to copy the floppy contents for modification.
        """
        self.log.info("Copying floppy contents for modification")
        shutil.copyfile(self.orig_floppy, self.output_floppy)

    def install(self, timeout=None, force=False):
        """
        Method to run the operating system installation.
        """
        if not force and os.access(self.jeos_filename, os.F_OK):
            self.log.info("Found cached JEOS, using it")
            oz.ozutil.copyfile_sparse(self.jeos_filename, self.diskimage)
            return self._generate_xml("hd", None)

        self.log.info("Running install for %s", self.tdl.name)

        fddev = self._InstallDev("floppy", self.output_floppy, "fda")

        if timeout is None:
            timeout = 1200

        self._wait_for_install_finish(self._generate_xml("fd", fddev), timeout)

        if self.cache_jeos:
            self.log.info("Caching JEOS")
            oz.ozutil.mkdir_p(self.jeos_cache_dir)
            oz.ozutil.copyfile_sparse(self.diskimage, self.jeos_filename)

        return self._generate_xml("hd", None)

    def _cleanup_floppy(self):
        """
        Method to cleanup the temporary floppy data.
        """
        self.log.info("Cleaning up floppy data")
        oz.ozutil.rmtree_and_sync(self.floppy_contents)

    def cleanup_install(self):
        """
        Method to cleanup the installation floppies.
        """
        self.log.info("Cleaning up after install")
        try:
            os.unlink(self.output_floppy)
        except:
            pass

        if not self.cache_original_media:
            try:
                os.unlink(self.orig_floppy)
            except Exception:
                pass



# Copyright (C) 2013-2017  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Linux installation
"""

import os
import re
import time

import libvirt

import oz.Guest
import oz.OzException


class LinuxCDGuest(oz.Guest.CDGuest):
    """
    Class for Linux installation.
    """
    def __init__(self, tdl, config, auto, output_disk, nicmodel, diskbus,
                 iso_allowed, url_allowed, macaddress):
        oz.Guest.CDGuest.__init__(self, tdl, config, auto, output_disk,
                                  nicmodel, None, None, diskbus, iso_allowed,
                                  url_allowed, macaddress)

    def _test_ssh_connection(self, guestaddr):
        """
        Internal method to test out the ssh connection before we try to use it.
        Under systemd, the IP address of a guest can come up and reportip can
        run before the ssh key is generated and sshd starts up.  This check
        makes sure that we allow an additional 30 seconds (1 second per ssh
        attempt) for sshd to finish initializing.
        """
        count = 30
        success = False
        while count > 0:
            try:
                self.log.debug("Testing ssh connection, try %d", count)
                start = time.time()
                self.guest_execute_command(guestaddr, 'ls', timeout=1)
                self.log.debug("Succeeded")
                success = True
                break
            except oz.ozutil.SubprocessException:
                # ensure that we spent at least one second before trying again
                end = time.time()
                if (end - start) < 1:
                    time.sleep(1 - (end - start))
                count -= 1

        if not success:
            self.log.debug("Failed to connect to ssh on running guest")
            raise oz.OzException.OzException("Failed to connect to ssh on running guest")

    def get_default_runlevel(self, g_handle):
        """
        Function to determine the default runlevel based on the /etc/inittab.
        """
        runlevel = "3"
        if g_handle.exists('/etc/inittab'):
            lines = g_handle.cat('/etc/inittab').split("\n")
            for line in lines:
                if re.match('id:', line):
                    runlevel = line.split(':')[1]
                    break

        return runlevel

    def guest_execute_command(self, guestaddr, command, timeout=10):
        """
        Method to execute a command on the guest and return the output.
        """
        # ServerAliveInterval protects against NAT firewall timeouts
        # on long-running commands with no output
        #
        # PasswordAuthentication=no prevents us from falling back to
        # keyboard-interactive password prompting
        #
        # -F /dev/null makes sure that we don't use the global or per-user
        # configuration files

        return oz.ozutil.subprocess_check_output(["ssh", "-i", self.sshprivkey,
                                                  "-F", "/dev/null",
                                                  "-o", "ServerAliveInterval=30",
                                                  "-o", "StrictHostKeyChecking=no",
                                                  "-o", "ConnectTimeout=" + str(timeout),
                                                  "-o", "UserKnownHostsFile=/dev/null",
                                                  "-o", "PasswordAuthentication=no",
                                                  "-o", "IdentitiesOnly yes",
                                                  "root@" + guestaddr, command],
                                                 printfn=self.log.debug)

    def guest_live_upload(self, guestaddr, file_to_upload, destination,
                          timeout=10):
        """
        Method to copy a file to the live guest.
        """
        self.guest_execute_command(guestaddr,
                                   "mkdir -p " + os.path.dirname(destination),
                                   timeout)

        # ServerAliveInterval protects against NAT firewall timeouts
        # on long-running commands with no output
        #
        # PasswordAuthentication=no prevents us from falling back to
        # keyboard-interactive password prompting
        #
        # -F /dev/null makes sure that we don't use the global or per-user
        # configuration files
        return oz.ozutil.subprocess_check_output(["scp", "-i", self.sshprivkey,
                                                  "-F", "/dev/null",
                                                  "-o", "ServerAliveInterval=30",
                                                  "-o", "StrictHostKeyChecking=no",
                                                  "-o", "ConnectTimeout=" + str(timeout),
                                                  "-o", "UserKnownHostsFile=/dev/null",
                                                  "-o", "PasswordAuthentication=no",
                                                  "-o", "IdentitiesOnly yes",
                                                  file_to_upload,
                                                  "root@" + guestaddr + ":" + destination],
                                                 printfn=self.log.debug)

    def _customize_files(self, guestaddr):
        """
        Method to upload the custom files specified in the TDL to the guest.
        """
        self.log.info("Uploading custom files")
        for name, fp in list(self.tdl.files.items()):
            # all of the self.tdl.files are named temporary files; we just need
            # to fetch the name out and have scp upload it
            self.guest_live_upload(guestaddr, fp.name, name)

    def _shutdown_guest(self, guestaddr, libvirt_dom):
        """
        Method to shutdown the guest (gracefully at first, then with prejudice).
        """
        if guestaddr is not None:
            # sometimes the ssh process gets disconnected before it can return
            # cleanly (particularly when the guest is running systemd).  If that
            # happens, ssh returns 255, guest_execute_command throws an
            # exception, and the guest is forcibly destroyed.  While this
            # isn't the end of the world, it isn't desirable.  To avoid
            # this, we catch any exception thrown by ssh during the shutdown
            # command and throw them away.  In the (rare) worst case, the
            # shutdown will not have made it to the guest and we'll have to wait
            # 90 seconds for wait_for_guest_shutdown to timeout and forcibly
            # kill the guest.
            try:
                self.guest_execute_command(guestaddr, 'shutdown -h now')
            except Exception:
                pass

            try:
                if not self._wait_for_guest_shutdown(libvirt_dom):
                    self.log.warning("Guest did not shutdown in time, going to kill")
                else:
                    libvirt_dom = None
            except Exception:
                self.log.warning("Failed shutting down guest, forcibly killing")

        if libvirt_dom is not None:
            try:
                libvirt_dom.destroy()
            except libvirt.libvirtError:
                # the destroy failed for some reason.  This can happen if
                # _wait_for_guest_shutdown times out, but the domain shuts
                # down before we get to destroy.  Check to make sure that the
                # domain is gone from the list of running domains; if so, just
                # continue on; if not, re-raise the error.
                for domid in self.libvirt_conn.listDomainsID():
                    if domid == libvirt_dom.ID():
                        raise

    def _collect_setup(self, libvirt_xml):  # pylint: disable=unused-argument
        """
        Default method to set the guest up for remote access.
        """
        raise oz.OzException.OzException("ICICLE generation and customization is not implemented for guest %s" % (self.tdl.distro))

    def _collect_teardown(self, libvirt_xml):  # pylint: disable=unused-argument
        """
        Method to reverse the changes done in _collect_setup.
        """
        raise oz.OzException.OzException("ICICLE generation and customization is not implemented for guest %s" % (self.tdl.distro))

    def _install_packages(self, guestaddr, packstr):  # pylint: disable=unused-argument
        """
        Internal method to install packages; expected to be overriden by
        child classes.
        """
        raise oz.OzException.OzException("Customization is not implemented for guest %s" % (self.tdl.distro))

    def _customize_repos(self, guestaddr):  # pylint: disable=unused-argument
        """
        Internal method to customize repositories; expected to be overriden by
        child classes.
        """
        raise oz.OzException.OzException("Customization is not implemented for guest %s" % (self.tdl.distro))

    def _remove_repos(self, guestaddr):  # pylint: disable=unused-argument
        """
        Internal method to remove repositories; expected to be overriden by
        child classes.
        """
        raise oz.OzException.OzException("Repository removal not implemented for guest %s" % (self.tdl.distro))

    def do_customize(self, guestaddr):
        """
        Method to customize by installing additional packages and files.
        """
        if not self.tdl.packages and not self.tdl.files and not self.tdl.commands:
            # no work to do, just return
            return

        self._customize_repos(guestaddr)

        for cmd in self.tdl.precommands:
            self.guest_execute_command(guestaddr, cmd.read())

        self.log.debug("Installing custom packages")
        packstr = ''
        for package in self.tdl.packages:
            packstr += '"' + package.name + '" '

        if packstr != '':
            self._install_packages(guestaddr, packstr)

        self._customize_files(guestaddr)

        self.log.debug("Running custom commands")
        for cmd in self.tdl.commands:
            self.guest_execute_command(guestaddr, cmd.read())

        self.log.debug("Removing non-persisted repos")
        self._remove_repos(guestaddr)

        self.log.debug("Syncing")
        self.guest_execute_command(guestaddr, 'sync')

    def do_icicle(self, guestaddr):
        """
        Default method to collect the package information and generate the
        ICICLE XML.
        """
        raise oz.OzException.OzException("ICICLE generation is not implemented for this guest type")

    def _internal_customize(self, libvirt_xml, action):
        """
        Internal method to customize and optionally generate an ICICLE for the
        operating system after initial installation.
        """
        # the "action" input is actually a tri-state:
        # action = "gen_and_mod" means to generate the icicle and to
        #          potentially make modifications
        # action = "gen_only" means to generate the icicle only, and not
        #          look at any modifications
        # action = "mod_only" means to not generate the icicle, but still
        #          potentially make modifications

        self.log.info("Customizing image")

        if not self.tdl.packages and not self.tdl.files and not self.tdl.commands:
            if action == "mod_only":
                self.log.info("No additional packages, files, or commands to install, and icicle generation not requested, skipping customization")
                return
            elif action == "gen_and_mod":
                # It is actually possible to get here with a "gen_and_mod"
                # action but a TDL that contains no real customizations.
                # In the "safe ICICLE" code below it is important to know
                # when we are truly in a "gen_only" state so we modify
                # the action here if we detect that ICICLE generation is the
                # only task to be done.
                self.log.debug("Asked to gen_and_mod but no mods are present - changing action to gen_only")
                action = "gen_only"

        # when doing an oz-install with -g, this isn't necessary as it will
        # just replace the port with the same port.  However, it is very
        # necessary when doing an oz-customize since the serial port might
        # not match what is specified in the libvirt XML
        modified_xml = self._modify_libvirt_xml_for_serial(libvirt_xml)

        if action == "gen_only" and self.safe_icicle_gen:
            # We are only generating ICICLE and the user has asked us to do
            # this without modifying the completed image by booting it.
            # Create a copy on write snapshot to use for ICICLE
            # generation - discard when finished
            cow_diskimage = self.diskimage + "-icicle-snap.qcow2"
            self._internal_generate_diskimage(force=True,
                                              backing_filename=self.diskimage,
                                              image_filename=cow_diskimage)
            modified_xml = self._modify_libvirt_xml_diskimage(modified_xml, cow_diskimage, 'qcow2')

        self._collect_setup(modified_xml)

        icicle = None
        try:
            libvirt_dom = self.libvirt_conn.createXML(modified_xml, 0)

            try:
                guestaddr = None
                guestaddr = self._wait_for_guest_boot(libvirt_dom)
                self._test_ssh_connection(guestaddr)

                if action == "gen_and_mod":
                    self.do_customize(guestaddr)
                    icicle = self.do_icicle(guestaddr)
                elif action == "gen_only":
                    icicle = self.do_icicle(guestaddr)
                elif action == "mod_only":
                    self.do_customize(guestaddr)
                else:
                    raise oz.OzException.OzException("Invalid customize action %s; this is a programming error" % (action))
            finally:
                if action == "gen_only" and self.safe_icicle_gen:
                    # if this is a gen_only and safe_icicle_gen, there is no
                    # reason to wait around for the guest to shutdown; we'll
                    # be removing the overlay file anyway.  Just destroy it
                    libvirt_dom.destroy()
                else:
                    self._shutdown_guest(guestaddr, libvirt_dom)
        finally:
            if action == "gen_only" and self.safe_icicle_gen:
                # no need to teardown because we simply discard the file
                # containing those changes
                os.unlink(cow_diskimage)
            else:
                self._collect_teardown(modified_xml)

        return icicle

    def customize(self, libvirt_xml):
        """
        Method to customize the operating system after installation.
        """
        return self._internal_customize(libvirt_xml, "mod_only")

    def customize_and_generate_icicle(self, libvirt_xml):
        """
        Method to customize and generate the ICICLE for an operating system
        after installation.  This is equivalent to calling customize() and
        generate_icicle() back-to-back, but is faster.
        """
        return self._internal_customize(libvirt_xml, "gen_and_mod")

    def generate_icicle(self, libvirt_xml):
        """
        Method to generate the ICICLE from an operating system after
        installation.  The ICICLE contains information about packages and
        other configuration on the diskimage.
        """
        return self._internal_customize(libvirt_xml, "gen_only")



# Copyright (C) 2010,2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2018  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Miscellaneous utility functions.
"""

import collections
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import errno
import fcntl
import ftplib
import gzip
import logging
import os
import random
import select
import shutil
import socket
import stat
import struct
import subprocess
import sys
import time
import urllib

import lxml.etree

import monotonic

import requests


def generate_full_auto_path(relative):
    """
    Function to find the absolute path to an unattended installation file.
    """
    # all of the automated installation paths are installed to $pkg_path/auto,
    # so we just need to find it and generate the right path here
    if relative is None:
        raise Exception("The relative path cannot be None")

    pkg_path = os.path.dirname(__file__)
    return os.path.abspath(os.path.join(pkg_path, "auto", relative))


def executable_exists(program):
    """
    Function to find out whether an executable exists in the PATH
    of the user.  If so, the absolute path to the executable is returned.
    If not, an exception is raised.
    """
    def is_exe(fpath):
        """
        Helper method to check if a file exists and is executable
        """
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    if program is None:
        raise Exception("Invalid program name passed")

    fpath, fname_unused = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    raise Exception("Could not find %s" % (program))


def write_bytes_to_fd(fd, buf):
    """
    Function to write all bytes in "buf" to "fd".  This handles both EINTR
    and short writes.
    """
    size = len(buf)
    offset = 0
    while size > 0:
        try:
            bytes_written = os.write(fd, buf[offset:])
            offset += bytes_written
            size -= bytes_written
        except OSError as err:
            # python's os.write() can raise an exception on EINTR, which
            # according to the man page can happen if a signal was
            # received before any data was written.  Therefore, we don't
            # need to update destlen or size, but just retry
            if err.errno == errno.EINTR:
                continue
            raise

    return offset


def read_bytes_from_fd(fd, num):
    """
    Function to read and return bytes from fd.  This handles the EINTR situation
    where no bytes were read before a signal happened.
    """
    read_done = False
    while not read_done:
        try:
            ret = os.read(fd, num)
            read_done = True
        except OSError as err:
            # python's os.read() can raise an exception on EINTR, which
            # according to the man page can happen if a signal was
            # received before any data was read.  In this case we need to retry
            if err.errno == errno.EINTR:
                continue
            raise

    return ret


def copyfile_sparse(src, dest):
    """
    Function to copy a file sparsely if possible.  The logic here is
    all taken from coreutils cp, specifically the 'sparse_copy' function.
    """
    if src is None:
        raise Exception("Source of copy cannot be None")
    if dest is None:
        raise Exception("Destination of copy cannot be None")

    if not os.path.exists(src):
        raise Exception("Source '%s' does not exist" % (src))

    if os.path.exists(dest) and os.path.samefile(src, dest):
        raise Exception("Source '%s' and dest '%s' are the same file" % (src, dest))

    base = os.path.dirname(dest)
    if not os.path.exists(base):
        mkdir_p(base)

    src_fd = os.open(src, os.O_RDONLY)

    try:
        dest_fd = os.open(dest, os.O_WRONLY | os.O_CREAT | os.O_TRUNC)

        try:
            sb = os.fstat(src_fd)

            # See io_blksize() in coreutils for an explanation of why 32*1024
            buf_size = max(32 * 1024, sb.st_blksize)

            size = sb.st_size
            destlen = 0
            while size != 0:
                buf = read_bytes_from_fd(src_fd, min(buf_size, size))
                if not buf:
                    break

                buflen = len(buf)
                if buf == '\0' * buflen:
                    os.lseek(dest_fd, buflen, os.SEEK_CUR)
                else:
                    write_bytes_to_fd(dest_fd, buf)

                destlen += buflen
                size -= buflen

            os.ftruncate(dest_fd, destlen)

        finally:
            os.close(dest_fd)
    finally:
        os.close(src_fd)


def bsd_split(line, digest_type):
    """
    Function to split a BSD-style checksum line into a filename and
    checksum.
    """
    current = len(digest_type)

    if line[current] == ' ':
        current += 1

    if line[current] != '(':
        return None, None

    current += 1

    # find end of filename.  The BSD 'md5' and 'sha1' commands do not escape
    # filenames, so search backwards for the last ')'
    file_end = line.rfind(')')
    if file_end == -1:
        # could not find the ending ), fail
        return None, None

    filename = line[current:file_end]

    line = line[(file_end + 1):]
    line = line.lstrip()

    if line[0] != '=':
        return None, None

    line = line[1:]

    line = line.lstrip()
    if line[-1] == '\n':
        line = line[:-1]

    return line, filename


def sum_split(line, digest_bits):
    """
    Function to split a normal Linux checksum line into a filename and
    checksum.
    """
    digest_hex_bytes = digest_bits // 4
    min_digest_line_length = digest_hex_bytes + 2 + 1  # length of hex message digest + blank and binary indicator (2 bytes) + minimum file length (1 byte)

    min_length = min_digest_line_length
    if line[0] == '\\':
        min_length = min_length + 1
    if len(line) < min_length:
        # if the line is too short, skip it
        return None, None

    if line[0] == '\\':
        current = digest_hex_bytes + 1
        hex_digest = line[1:current]
        escaped_filename = True
    else:
        current = digest_hex_bytes
        hex_digest = line[0:current]
        escaped_filename = False

    # if the digest is not immediately followed by a white space, it is an
    # error
    if line[current] != ' ' and line[current] != '\t':
        return None, None

    current += 1
    # if the whitespace is not immediately followed by another space or a *,
    # it is an error
    if line[current] != ' ' and line[current] != '*':
        return None, None

    current += 1

    filename = line[current:]
    if line[-1] == '\n':
        filename = line[current:-1]

    if escaped_filename:
        # FIXME: a \0 is not allowed in the sum file format, but
        # string_escape allows it.  We'd probably have to implement our
        # own codec to fix this
        if sys.version_info.major == 2:
            filename = filename.decode('string_escape')
        else:
            filename = filename.encode('utf-8').decode('unicode_escape')

    return hex_digest, filename


def get_sum_from_file(sumfile, file_to_find, digest_bits, digest_type):
    """
    Function to get a checksum digest out of a checksum file given a
    filename.
    """
    retval = None

    with open(sumfile, 'r') as f:
        for line in f:
            # remove any leading whitespace
            line = line.lstrip()

            # ignore blank lines
            if not line:
                continue

            # ignore comment lines
            if line[0] == '#':
                continue

            if line.startswith(digest_type):
                # OK, if it starts with a string of ["MD5", "SHA1", "SHA256"], then
                # this is a BSD-style sumfile
                hex_digest, filename = bsd_split(line, digest_type)
            else:
                # regular sumfile
                hex_digest, filename = sum_split(line, digest_bits)

            if hex_digest is None or filename is None:
                continue

            if filename == file_to_find:
                retval = hex_digest
                break

    return retval


def get_md5sum_from_file(sumfile, file_to_find):
    """
    Function to get an MD5 checksum out of a checksum file given a filename.
    """
    return get_sum_from_file(sumfile, file_to_find, 128, "MD5")


def get_sha1sum_from_file(sumfile, file_to_find):
    """
    Function to get a SHA1 checksum out of a checksum file given a filename.
    """
    return get_sum_from_file(sumfile, file_to_find, 160, "SHA1")


def get_sha256sum_from_file(sumfile, file_to_find):
    """
    Function to get a SHA256 checksum out of a checksum file given a
    filename.
    """
    return get_sum_from_file(sumfile, file_to_find, 256, "SHA256")


def string_to_bool(instr):
    """
    Function to take a string and determine whether it is True, Yes, False,
    or No.  It takes a single argument, which is the string to examine.

    Returns True if instr is "Yes" or "True", False if instr is "No"
    or "False", and None otherwise.
    """
    if instr is None:
        raise Exception("Input string was None!")
    lower = instr.lower()
    if lower == 'no' or lower == 'false':
        return False
    if lower == 'yes' or lower == 'true':
        return True
    return None


def generate_macaddress():
    """
    Function to generate a random MAC address.
    """
    mac = [0x52, 0x54, 0x00, random.randint(0x00, 0xff),
           random.randint(0x00, 0xff), random.randint(0x00, 0xff)]
    return ':'.join(["%02x" % x for x in mac])


class SubprocessException(Exception):
    """
    Class for subprocess exceptions.  In addition to a error message, it
    also has a retcode member that has the returncode from the command.
    """
    def __init__(self, msg, retcode):
        Exception.__init__(self, msg)
        self.retcode = retcode


def subprocess_check_output(*popenargs, **kwargs):
    """
    Function to call a subprocess and gather the output.
    """
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')
    if 'stderr' in kwargs:
        raise ValueError('stderr argument not allowed, it will be overridden.')

    printfn = None
    if 'printfn' in kwargs:
        printfn = kwargs['printfn']
        del kwargs['printfn']

    executable_exists(popenargs[0][0])

    process = subprocess.Popen(stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               *popenargs, **kwargs)

    poller = select.poll()
    select_POLLIN_POLLPRI = select.POLLIN | select.POLLPRI
    poller.register(process.stdout.fileno(), select_POLLIN_POLLPRI)
    poller.register(process.stderr.fileno(), select_POLLIN_POLLPRI)

    stdout = ''
    stderr = ''
    retcode = process.poll()
    while retcode is None:
        start = time.time()
        try:
            ready = poller.poll(1000)
        except select.error as e:
            if e.args[0] == errno.EINTR:
                continue
            raise

        for fd, mode in ready:
            if mode & select_POLLIN_POLLPRI:
                data = os.read(fd, 4096)
                if not data:
                    poller.unregister(fd)
                else:
                    data = data.decode('utf-8')
                    if printfn is not None:
                        printfn(data)
                    if fd == process.stdout.fileno():
                        stdout += data
                    else:
                        stderr += data
            else:
                # Ignore hang up or errors.
                poller.unregister(fd)

        end = time.time()
        if (end - start) < 1:
            time.sleep(1 - (end - start))
        retcode = process.poll()

    tmpout, tmperr = process.communicate()

    tmpout = tmpout.decode('utf-8')
    tmperr = tmperr.decode('utf-8')

    stdout += tmpout
    stderr += tmperr
    if printfn is not None:
        printfn(tmperr)
        printfn(tmpout)

    if retcode:
        cmd = ' '.join(*popenargs)
        raise SubprocessException("'%s' failed(%d): %s" % (cmd, retcode, stderr + stdout), retcode)

    return (stdout, stderr, retcode)


def mkdir_p(path):
    """
    Function to make a directory and all intermediate directories as
    necessary.  The functionality differs from os.makedirs slightly, in
    that this function does *not* raise an error if the directory already
    exists.
    """
    if path is None:
        raise Exception("Path cannot be None")

    if path == '':
        # this can happen if the user did something like call os.path.dirname()
        # on a file without directories.  Since os.makedirs throws an exception
        # in that case, check for it here and allow it.
        return

    try:
        os.makedirs(path)
    except OSError as err:
        if err.errno != errno.EEXIST or not os.path.isdir(path):
            raise


def copytree_merge(src, dst, symlinks=False, ignore=None):
    """
    Function to copy an entire directory recursively. The functionality
    differs from shutil.copytree, in that this function does *not* raise
    an exception if the directory already exists.
    It is based on: http://docs.python.org/2.7/library/shutil.html#copytree-example
    """
    names = os.listdir(src)
    if ignore is not None:
        ignored_names = ignore(src, names)
    else:
        ignored_names = set()

    mkdir_p(dst)
    errors = []
    for name in names:
        if name in ignored_names:
            continue
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if symlinks and os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                copytree_merge(srcname, dstname, symlinks, ignore)
            else:
                shutil.copy2(srcname, dstname)
            # FIXME: What about devices, sockets etc.?
        except (IOError, os.error) as why:
            errors.append((srcname, dstname, str(why)))
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
    try:
        shutil.copystat(src, dst)
    except OSError as why:
        errors.extend((src, dst, str(why)))
    if errors:
        raise shutil.Error(errors)


def copy_modify_file(inname, outname, subfunc):
    """
    Function to copy a file from inname to outname, passing each line
    through subfunc first.  subfunc is expected to be a method that takes
    a single argument in (the next line), and returns a string to be
    written to the output file after modification (if any).
    """
    if inname is None:
        raise Exception("input filename is None")
    if outname is None:
        raise Exception("output filename is None")
    if subfunc is None:
        raise Exception("subfunction is None")
    if not isinstance(subfunc, collections.Callable):
        raise Exception("subfunction is not callable")

    infile = open(inname, 'r')
    outfile = open(outname, 'w')

    for line in infile:
        outfile.write(subfunc(line))

    infile.close()
    outfile.close()


def write_cpio(inputdict, outputfile):
    """
    Function to write a CPIO archive in the "New ASCII Format".  The
    inputlist is a dictionary of files to put in the archive, where the
    dictionary key is the path to the file on the local filesystem and the
    dictionary value is the location that the file should have in the cpio
    archive.  The outputfile is the location of the final cpio archive that
    will be written.
    """
    if inputdict is None:
        raise Exception("input dictionary was None")
    if outputfile is None:
        raise Exception("output file was None")

    outf = open(outputfile, "w")

    try:
        for inputfile, destfile in list(inputdict.items()):
            inf = open(inputfile, 'r')
            st = os.fstat(inf.fileno())

            # 070701 is the magic for new CPIO (newc in cpio parlance)
            outf.write("070701")
            # inode (really just needs to be unique)
            outf.write("%08x" % (st[stat.ST_INO]))
            # mode
            outf.write("%08x" % (st[stat.ST_MODE]))
            # uid is 0
            outf.write("00000000")
            # gid is 0
            outf.write("00000000")
            # nlink (always a single link for a single file)
            outf.write("00000001")
            # mtime
            outf.write("%08x" % (st[stat.ST_MTIME]))
            # filesize
            outf.write("%08x" % (st[stat.ST_SIZE]))
            # devmajor
            outf.write("%08x" % (os.major(st[stat.ST_DEV])))
            # dev minor
            outf.write("%08x" % (os.minor(st[stat.ST_DEV])))
            # rdevmajor (always 0)
            outf.write("00000000")
            # rdevminor (always 0)
            outf.write("00000000")
            # namesize (the length of the name plus 1 for the NUL padding)
            outf.write("%08x" % (len(destfile) + 1))
            # check (always 0)
            outf.write("00000000")
            # write the name of the inputfile minus the leading /
            stripped = destfile.lstrip('/')
            outf.write(stripped)

            # we now need to write sentinel NUL byte(s).  We need to make the
            # header (110 bytes) plus the filename, plus the sentinel a
            # multiple of 4 bytes.  Note that we always need at *least* one NUL,
            # so if it is exactly a multiple of 4 we need to write 4 NULs
            outf.write("\x00" * (4 - ((110 + len(stripped)) % 4)))

            # now write the data from the input file
            outf.writelines(inf)
            inf.close()

            # we now need to write out NUL byte(s) to make it a multiple of 4.
            # note that unlike the name, we do *not* have to have any NUL bytes,
            # so if it is already aligned on 4 bytes do nothing
            remainder = st[stat.ST_SIZE] % 4
            if remainder != 0:
                outf.write("\x00" * (4 - remainder))

        # now that we have written all of the file entries, write the trailer
        outf.write("070701")
        # zero inode
        outf.write("00000000")
        # zero mode
        outf.write("00000000")
        # zero uid
        outf.write("00000000")
        # zero gid
        outf.write("00000000")
        # one nlink
        outf.write("00000001")
        # zero mtime
        outf.write("00000000")
        # zero filesize
        outf.write("00000000")
        # zero devmajor
        outf.write("00000000")
        # zero devminor
        outf.write("00000000")
        # zero rdevmajor
        outf.write("00000000")
        # zero rdevminor
        outf.write("00000000")
        # 0xB namesize
        outf.write("0000000B")
        # zero check
        outf.write("00000000")
        # trailer
        outf.write("TRAILER!!!")

        # finally, we need to pad to the closest 512 bytes
        outf.write("\x00" * (512 - (outf.tell() % 512)))
    except:
        os.unlink(outputfile)
        raise

    outf.close()


def config_get_key(config, section, key, default):
    """
    Function to retrieve config parameters out of the config file.
    """
    if config is not None and config.has_section(section) and config.has_option(section, key):
        return config.get(section, key)
    return default


def config_get_boolean_key(config, section, key, default):
    """
    Function to retrieve boolean config parameters out of the config file.
    """
    value = config_get_key(config, section, key, None)
    if value is None:
        return default

    retval = string_to_bool(value)
    if retval is None:
        raise Exception("Configuration parameter '%s' must be True, Yes, False, or No" % (key))

    return retval


def config_get_path(config, section, key, default):
    """
    Function to get an user-expanded path out of the config file at
    the passed in section and key.  If the value is not in the config
    file, then the default value is returned.  If the expanded path is
    not absolute, an error is raised.
    """
    path = os.path.expanduser(config_get_key(config, section, key, default))
    if not os.path.isabs(path):
        raise Exception("Config key '%s' must have an absolute path" % (key))
    return path


def rmtree_and_sync(directory):
    """
    Function to remove a directory tree and do an fsync afterwards.  Because
    the removal of the directory tree can cause a lot of metadata updates, it
    can cause a lot of disk activity.  By doing the fsync, we ensure that any
    metadata updates caused by us will not cause subsequent steps to fail.  This
    cannot help if the system is otherwise very busy, but it does ensure that
    the problem is not self-inflicted.
    """
    try:
        shutil.rmtree(directory)
        fd = os.open(os.path.dirname(directory), os.O_RDONLY)
        try:
            os.fsync(fd)
        finally:
            os.close(fd)
    except OSError as err:
        if err.errno == 2:
            pass
        else:
            raise


def parse_config(config_file):
    """
    Function to parse the configuration file.  If the passed in config_file is
    None, then the default configuration file is used.
    """
    config = configparser.SafeConfigParser()
    if config_file is not None:
        # If the config_file passed in is not None, then we want to try to read
        # that config file (after expanding it).  If that config file doesn't
        # exist, we want to throw an error (which is why we use readfp here).
        config.readfp(open(os.path.expanduser(config_file)))
    else:
        # The config file was not passed in, so we want to use one of the
        # defaults.  First we check to see if a ~/.oz/oz.cfg exists; if it does,
        # we use that.  Otherwise we fall back to the system-wide version in
        # /etc/oz/oz.cfg.  If neither of those exist, we don't throw an error
        # but instead let Oz pick sane defaults internally.
        parsed = config.read(os.path.expanduser("~/.oz/oz.cfg"))
        if not parsed and os.geteuid() == 0:
            config.read("/etc/oz/oz.cfg")

    return config


def default_output_dir():
    """
    Function to get the default path to the output directory.
    """
    if os.geteuid() == 0:
        return "/var/lib/libvirt/images"
    return "~/.oz/images"


def default_data_dir():
    """
    Function to get the default path to the data directory.
    """
    if os.geteuid() == 0:
        return "/var/lib/oz"
    return "~/.oz"


def default_sshprivkey():
    """
    Function to get the default path to the SSH private key.
    """
    if os.geteuid() == 0:
        return "/etc/oz/id_rsa-icicle-gen"
    return "~/.oz/id_rsa-icicle-gen"


def default_screenshot_dir():
    """
    Function to get the default path to the screenshot directory. The directory
    is generated relative to the default data directory.
    """
    return os.path.join(default_data_dir(), "screenshots")


class LocalFileAdapter(requests.adapters.BaseAdapter):
    '''
    This class implements an adapter for requests so we can properly deal with file://
    local files.
    '''
    @staticmethod
    def _chkpath(method, path):
        """Return an HTTP status for the given filesystem path."""
        if method.lower() in ('put', 'delete'):
            return 501, "Not Implemented"
        elif method.lower() not in ('get', 'head', 'post'):
            return 405, "Method Not Allowed"
        elif os.path.isdir(path):
            return 400, "Path Not A File"
        elif not os.path.isfile(path):
            return 404, "File Not Found"
        elif not os.access(path, os.R_OK):
            return 403, "Access Denied"
        return 200, "OK"

    def send(self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None):
        """Return the file specified by the given request

        @type req: C{PreparedRequest}
        @todo: Should I bother filling `response.headers` and processing
               If-Modified-Since and friends using `os.stat`?
        """
        if sys.version_info.major == 2:
            path = os.path.normcase(os.path.normpath(urllib.url2pathname(request.path_url)))  # pylint: disable=no-member
        else:
            path = os.path.normcase(os.path.normpath(urllib.request.url2pathname(request.path_url)))
        response = requests.Response()

        response.status_code, response.reason = self._chkpath(request.method, path)
        if response.status_code == 200 and request.method.lower() != 'head':
            try:
                response.raw = open(path, 'rb')
            except (OSError, IOError) as err:
                response.status_code = 500
                response.reason = str(err)

        if isinstance(request.url, bytes):
            response.url = request.url.decode('utf-8')
        else:
            response.url = request.url

        response.headers['Content-Length'] = os.path.getsize(path)
        response.headers['Accept-Ranges'] = 'bytes'
        response.headers['Redirect-URL'] = request.url
        response.request = request
        response.connection = self

        return response

    def close(self):
        pass


def http_get_header(url, redirect=True):
    """
    Function to get the HTTP headers from a URL.  The available headers will be
    returned in a dictionary.  If redirect=True (the default), then this
    function will automatically follow http redirects through to the final
    destination, entirely transparently to the caller.  If redirect=False, then
    this function will follow http redirects through to the final destination,
    and also store that information in the 'Redirect-URL' key.  Note that
    'Redirect-URL' will always be None in the redirect=True case, and may be
    None in the redirect=True case if no redirects were required.
    """
    with requests.Session() as requests_session:
        requests_session.mount('file://', LocalFileAdapter())
        response = requests_session.head(url, allow_redirects=redirect, stream=True, timeout=10)
        info = response.headers
        info['HTTP-Code'] = response.status_code
        if not redirect:
            info['Redirect-URL'] = response.headers.get('Location')
        else:
            info['Redirect-URL'] = None

    return info


def http_download_file(url, fd, show_progress, logger):
    """
    Function to download a file from url to file descriptor fd.
    """
    with requests.Session() as requests_session:
        requests_session.mount('file://', LocalFileAdapter())
        response = requests_session.get(url, stream=True, allow_redirects=True,
                                        headers={'Accept-Encoding': ''})
        file_size = int(response.headers.get('Content-Length'))
        chunk_size = 10 * 1024 * 1024
        done = 0
        for chunk in response.iter_content(chunk_size):
            write_bytes_to_fd(fd, chunk)
            done += len(chunk)
            if show_progress:
                logger.debug("%dkB of %dkB" % (done / 1024, file_size / 1024))


def ftp_download_directory(server, username, password, basepath, destination, port=None):
    """
    Function to recursively download an entire directory structure over FTP.
    """
    ftp = ftplib.FTP()
    ftp.connect(server, port)
    ftp.login(username, password)

    def _recursive_ftp_download(sourcepath):
        """
        Function to iterate and download a remote ftp folder
        """
        original_dir = ftp.pwd()
        try:
            ftp.cwd(sourcepath)
        except ftplib.error_perm:
            relativesourcepath = os.path.relpath(sourcepath, basepath)
            destinationpath = os.path.join(destination, relativesourcepath)
            if not os.path.exists(os.path.dirname(destinationpath)):
                os.makedirs(os.path.dirname(destinationpath))
            ftp.retrbinary("RETR " + sourcepath, open(destinationpath, "wb").write)
            return

        names = ftp.nlst()
        for name in names:
            _recursive_ftp_download(os.path.join(sourcepath, name))

        ftp.cwd(original_dir)

    _recursive_ftp_download(basepath)
    ftp.close()


def _gzip_file(inputfile, outputfile, outputmode):
    """
    Internal function to gzip the input file and place it in the outputfile.
    If the outputmode is 'ab', then the input file will be appended to the
    output file, and if the outputmode is 'wb' then the input file will be
    written over the output file.
    """
    with open(inputfile, 'rb') as f:
        gzf = gzip.GzipFile(outputfile, mode=outputmode)
        gzf.writelines(f)
        gzf.close()


def gzip_append(inputfile, outputfile):
    """
    Function to gzip and append the data from inputfile onto output file.
    """
    _gzip_file(inputfile, outputfile, 'ab')


def gzip_create(inputfile, outputfile):
    """
    Function to gzip the data from inputfile and place it into outputfile,
    overwriting any existing data in outputfile.
    """
    try:
        _gzip_file(inputfile, outputfile, 'wb')
    except:
        # since we created the output file, we should clean it up
        if os.access(outputfile, os.F_OK):
            os.unlink(outputfile)
        raise


def check_qcow_size(filename):
    """
    Function to detect if an image is in qcow format.  If it is, return the size
    of the underlying disk image.  If it isn't, return None.
    """

    # For interested parties, this is the QCOW header struct in C
    # struct qcow_header {
    #    uint32_t magic;
    #    uint32_t version;
    #    uint64_t backing_file_offset;
    #    uint32_t backing_file_size;
    #    uint32_t cluster_bits;
    #    uint64_t size; /* in bytes */
    #    uint32_t crypt_method;
    #    uint32_t l1_size;
    #    uint64_t l1_table_offset;
    #    uint64_t refcount_table_offset;
    #    uint32_t refcount_table_clusters;
    #    uint32_t nb_snapshots;
    #    uint64_t snapshots_offset;
    # };

    # And in Python struct format string-ese
    qcow_struct = ">IIQIIQIIQQIIQ"  # > means big-endian
    qcow_magic = 0x514649FB  # 'Q' 'F' 'I' 0xFB

    f = open(filename, "r")
    pack = f.read(struct.calcsize(qcow_struct))
    f.close()

    unpack = struct.unpack(qcow_struct, pack)

    if unpack[0] == qcow_magic:
        return unpack[5]
    return None


def recursively_add_write_bit(inputdir):
    """
    Function to walk a directory tree, adding the write it to every file
    and directory.  This is mostly useful right before deleting a tree of
    files extracted from an ISO, since those were all read-only to begin
    with.
    """
    for dirpath, dirnames_unused, filenames in os.walk(inputdir):
        # If the path is a symlink, and it is an absolute symlink, this would
        # attempt to change the permissions of the *host* file, not the
        # file that is relative to here.  That is no good, and could be a
        # security problem if Oz is being run as root.  We skip all paths that
        # are symlinks; what they point to will be changed later on.
        if os.path.islink(dirpath):
            continue

        os.chmod(dirpath, os.stat(dirpath).st_mode | stat.S_IWUSR)
        for name in filenames:
            fullpath = os.path.join(dirpath, name)

            # we have the same guard for symlinks as above, for the same reason
            if os.path.islink(fullpath):
                continue

            try:
                # if there are broken symlinks in the ISO,
                # then the below might fail.  This probably
                # isn't fatal, so just allow it and go on
                os.chmod(fullpath, os.stat(fullpath).st_mode | stat.S_IWUSR)
            except OSError as err:
                if err.errno != errno.ENOENT:
                    raise


def find_uefi_firmware(arch):
    '''
    A function to find the UEFI firmware file that corresponds to a certain
    architecture.
    '''
    # Yuck.  Finding the UEFI firmware to start certain guests (like aarch64)
    # is a really nasty process.  While slightly out of date, this blog post
    # describes the mess: http://blog.wikichoon.com/2016/01/uefi-support-in-virt-install-and-virt.html
    # Here, we replicate what libguestfs is doing here, which is to essentially
    # hardcode paths where UEFI firmware can be found on popular distributions.
    # I verified that these files exist on both Fedora/RHEL and Ubuntu.
    # Hopefully there will be a nicer way to do this in the future.
    class UEFI(object):
        '''
        A private class to hold the path to the loader and the path to the
        NVRAM file.
        '''
        def __init__(self, loader, nvram):
            self.loader = loader
            self.nvram = nvram

        def exists(self):
            '''
            A method that returns True if both the loader and the NVRAM files
            exist, False otherwise.
            '''
            if os.path.exists(self.loader) and os.path.exists(self.nvram):
                return True
            return False

    if arch in ['i386', 'i486', 'i586', 'i686']:
        uefi_list = [UEFI('/usr/share/edk2.git/ovmf-ia32/OVMF_CODE-pure-efi.fd',
                          '/usr/share/edk2.git/ovmf-ia32/OVMF_VARS-pure-efi.fd')]
    elif arch in ['x86_64']:
        uefi_list = [UEFI('/usr/share/OVMF/OVMF_CODE.fd',
                          '/usr/share/OVMF/OVMF_VARS.fd'),
                     UEFI('/usr/share/edk2.git/ovmf-x64/OVMF_CODE-pure-efi.fd',
                          '/usr/share/edk2.git/ovmf-x64/OVMF_VARS-pure-efi.fd')]
    elif arch in ['aarch64']:
        uefi_list = [UEFI('/usr/share/AAVMF/AAVMF_CODE.fd',
                          '/usr/share/AAVMF/AAVMF_VARS.fd'),
                     UEFI('/usr/share/edk2/aarch64/QEMU_EFI-pflash.raw',
                          '/usr/share/edk2/aarch64/vars-template-pflash.raw'),
                     UEFI('/usr/share/edk2.git/aarch64/QEMU_EFI-pflash.raw',
                          '/usr/share/edk2.git/aarch64/vars-template-pflash.raw')]
    else:
        raise Exception("Invalid arch for UEFI firmware")

    for uefi in uefi_list:
        if uefi.exists():
            return uefi.loader, uefi.nvram

    raise Exception("UEFI firmware is not installed!")


def open_locked_file(filename):
    """
    A function to open and lock a file.  Returns a file descriptor referencing
    the open and locked file.
    """
    outdir = os.path.dirname(filename)
    mkdir_p(outdir)

    fd = os.open(filename, os.O_RDWR | os.O_CREAT)

    try:
        fcntl.lockf(fd, fcntl.LOCK_EX)
    except:
        os.close(fd)
        raise

    return (fd, outdir)


def lxml_subelement(root, name, text=None, attributes=None):
    """
    Function to add a new element to an LXML tree, optionally include text
    and a dictionary of attributes.
    """
    tmp = lxml.etree.SubElement(root, name)
    if text is not None:
        tmp.text = text
    if attributes is not None:
        for k, v in attributes.items():
            tmp.set(k, v)
    return tmp


def timed_loop(max_time, cb, msg, cb_arg=None):
    '''
    A function to deal with waiting for an event to occur.  Given a
    maximum time to wait, a callback, and a message, it will wait until the maximum
    time for the event to occur.  Each time through the loop, it will do the following:

    1.  Check to see if it has been at least 10 seconds since it last logged.  If so, it
        will log right now.
    2.  Call the callback to check for the event.  If the callback returns True, the
        loop quits immediately.  If it returns False, go on to step 3.
    3.  Sleep for the portion of 1 second that was not taken up by the callback.

    If the event occurred (the callback returned True), then this function returns
    True.  If we timed out while waiting for the event to occur, this function returns
    False.
    '''
    log = logging.getLogger('%s' % (__name__))
    now = monotonic.monotonic()
    end = now + max_time
    next_print = now
    while now < end:
        now = monotonic.monotonic()
        if now >= next_print:
            left = int(end) - int(now)
            if left < 0:
                left = 0
            log.debug("%s, %d/%d", msg, left, max_time)
            next_print = now + 10

            if cb(cb_arg):
                return True

        # It's possible that the callback took longer than one second.
        # In that case, just skip our sleep altogether in an attempt to
        # catch up.
        sleep_time = 1.0 - (monotonic.monotonic() - now)
        if sleep_time > 0:
            # Otherwise, sleep for a time.  Note that we try to maintain
            # on our starting boundary, so we'll sleep less than a second
            # here almost always.
            time.sleep(sleep_time)

    return False


def get_free_port():
    """
    A function to find a free TCP port on the host.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # Bind to port 0 which will use a free socket to listen to.
    sock.bind(("", 0))
    listen_port = sock.getsockname()[1]
    # Close the socket to free it up for libvirt
    sock.close()

    return listen_port



# Copyright (C) 2017  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Helper class for managing a GuestFS connection
"""

import functools
import logging

import guestfs

import lxml.etree

import oz.OzException


class GuestFS(object):
    '''
    A wrapper class over guestfs to make some common operations easier.
    '''
    def __init__(self, input_disk, input_disk_type):
        self.log = logging.getLogger(__name__)

        self.g_handle = guestfs.GuestFS(python_return_dict=True)

        self.log.debug("Adding disk image %s", input_disk)
        # NOTE: we use "add_drive_opts" here so we can specify the type
        # of the diskimage.  Otherwise it might be possible for an attacker
        # to fool libguestfs with a specially-crafted diskimage that looks
        # like a qcow2 disk (thanks to rjones for the tip)
        self.g_handle.add_drive_opts(input_disk, format=input_disk_type)

        self.log.debug("Launching guestfs")
        self.g_handle.launch()

    def create_msdos_partition_table(self):
        '''
        A method to create a new msdos partition table on a disk.
        '''
        devices = self.g_handle.list_devices()
        self.g_handle.part_init(devices[0], "msdos")
        self.g_handle.part_add(devices[0], 'p', 1, 2)

    def mount_partitions(self):
        '''
        A method to mount existing partitions on a disk inside of guestfs.
        '''
        self.log.debug("Inspecting guest OS")
        roots = self.g_handle.inspect_os()

        if not roots:
            raise oz.OzException.OzException("No operating systems found on the disk")

        self.log.debug("Getting mountpoints")
        already_mounted = {}
        for root in roots:
            self.log.debug("Root device: %s", root)

            # the problem here is that the list of mountpoints returned by
            # inspect_get_mountpoints is in no particular order.  So if the
            # diskimage contains /usr and /usr/local on different devices,
            # but /usr/local happened to come first in the listing, the
            # devices would get mapped improperly.  The clever solution here is
            # to sort the mount paths by length; this will ensure that they
            # are mounted in the right order.  Thanks to rjones for the hint,
            # and the example code that comes from the libguestfs.org python
            # example page.
            mps = self.g_handle.inspect_get_mountpoints(root)

            for device in sorted(mps.keys(), key=len):
                try:
                    # Here we check to see if the device was already mounted.
                    # If it was, we skip over this mountpoint and go to the
                    # next one.  This can happen, for instance, on btrfs volumes
                    # with snapshots.  In that case, we'll always take the
                    # "original" backing filesystem, and not the snapshots,
                    # which seems to work in practice.
                    if already_mounted[device] == mps[device]:
                        continue
                except KeyError:
                    # If we got a KeyError exception, we know that we haven't
                    # yet mounted this filesystem, so continue on.
                    pass

                try:
                    self.g_handle.mount_options('', mps[device], device)
                    already_mounted[device] = mps[device]
                except Exception:
                    if device == '/':
                        # If we cannot mount root, we may as well give up
                        raise
                    else:
                        # some custom guests may have fstab content with
                        # "nofail" as a mount option.  For example, images
                        # built for EC2 with ephemeral mappings.  These
                        # fail at this point.  Allow things to continue.
                        # Profound failures will trigger later on during
                        # the process.
                        self.log.warning("Unable to mount (%s) on (%s) - trying to continue", mps[device], device)

    def remove_if_exists(self, path):
        """
        Method to remove a file if it exists in the disk image.
        """
        if self.g_handle.exists(path):
            self.g_handle.rm_rf(path)

    def move_if_exists(self, orig_path, replace_path):
        """
        Method to move a file if it exists in the disk image.
        """
        if self.g_handle.exists(orig_path):
            self.g_handle.mv(orig_path, replace_path)

    def path_backup(self, orig):
        """
        Method to backup a file in the disk image.
        """
        self.move_if_exists(orig, orig + ".ozbackup")

    def path_restore(self, orig):
        """
        Method to restore a backup file in the disk image.
        """
        backup = orig + ".ozbackup"
        self.remove_if_exists(orig)
        self.move_if_exists(backup, orig)

    def exists(self, filename):
        '''
        A passthrough method for the guestfs functionality of "exists".
        '''
        return self.g_handle.exists(filename)

    def rm(self, filename):
        '''
        A passthrough method for the guestfs functionality of "rm".
        '''
        return self.g_handle.rm(filename)

    def glob_expand(self, glob):
        '''
        A passthrough method for the guestfs functionality of "glob_expand".
        '''
        return self.g_handle.glob_expand(glob)

    def mkdir(self, directory):
        '''
        A passthrough method for the guestfs functionality of "mkdir".
        '''
        return self.g_handle.mkdir(directory)

    def ln_sf(self, src, dst):
        '''
        A passthrough method for the guestfs functionality of "ln_sf".
        '''
        return self.g_handle.ln_sf(src, dst)

    def chmod(self, mode, fname):
        '''
        A passthrough method for the guestfs functionality of "chmod".
        '''
        return self.g_handle.chmod(mode, fname)

    def cat(self, fname):
        '''
        A passthrough method for the guestfs functionality of "cat".
        '''
        return self.g_handle.cat(fname)

    def upload(self, src, dest):
        '''
        A passthrough method for the guestfs functionalityh of "upload".
        '''
        return self.g_handle.upload(src, dest)

    def cleanup(self):
        '''
        A method to cleanup after finishing with a guestfs handle.
        '''
        self.log.info("Cleaning up guestfs handle")
        self.log.debug("Syncing")
        self.g_handle.sync()

        self.log.debug("Unmounting all")
        self.g_handle.umount_all()
        self.g_handle.kill_subprocess()
        self.g_handle.close()


def GuestFSLibvirtFactory(libvirt_xml, libvirt_conn):
    '''
    A factory function for getting a GuestFS object from a libvirt XML and a connection.
    '''
    log = logging.getLogger(__name__)

    input_doc = lxml.etree.fromstring(libvirt_xml)
    namenode = input_doc.xpath('/domain/name')
    if len(namenode) != 1:
        raise oz.OzException.OzException("invalid libvirt XML with no name")
    input_name = namenode[0].text
    disks = input_doc.xpath('/domain/devices/disk')
    if len(disks) != 1:
        log.warning("Oz given a libvirt domain with more than 1 disk; using the first one parsed")
    source = disks[0].xpath('source')
    if len(source) != 1:
        raise oz.OzException.OzException("invalid <disk> entry without a source")
    input_disk = source[0].get('file')
    driver = disks[0].xpath('driver')
    if not driver:
        input_disk_type = 'raw'
    elif len(driver) == 1:
        input_disk_type = driver[0].get('type')
    else:
        raise oz.OzException.OzException("invalid <disk> entry without a driver")

    for domid in libvirt_conn.listDomainsID():
        try:
            doc = lxml.etree.fromstring(libvirt_conn.lookupByID(domid).XMLDesc(0))
        except Exception:
            log.debug("Could not get XML for domain ID (%s) - it may have disappeared (continuing)",
                      domid)
            continue

        namenode = doc.xpath('/domain/name')
        if len(namenode) != 1:
            # hm, odd, a domain without a name?
            raise oz.OzException.OzException("Saw a domain without a name, something weird is going on")
        if input_name == namenode[0].text:
            raise oz.OzException.OzException("Cannot setup ICICLE generation on a running guest")
        disks = doc.xpath('/domain/devices/disk')
        if len(disks) < 1:
            # odd, a domain without a disk, but don't worry about it
            continue
        for guestdisk in disks:
            for source in guestdisk.xpath("source"):
                # FIXME: this will only work for files; we can make it work
                # for other things by following something like:
                # http://git.annexia.org/?p=libguestfs.git;a=blob;f=src/virt.c;h=2c6be3c6a2392ab8242d1f4cee9c0d1445844385;hb=HEAD#l169
                filename = str(source.get('file'))
                if filename == input_disk:
                    raise oz.OzException.OzException("Cannot setup ICICLE generation on a running disk")

    return GuestFS(input_disk, input_disk_type)



# Copyright (C) 2010,2011  Chris Lalancette <clalance@redhat.com>
# Copyright (C) 2012-2018  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Template Description Language (TDL)
"""

import base64
import os
import re
import sys
import tempfile
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import lxml.etree

import oz.OzException
import oz.ozutil


def _xml_get_value(doc, xmlstring, component, optional=False):
    """
    Function to get the contents from an XML node.  It takes 4 arguments:

    doc       - The lxml.etree document to get a value from.
    xmlstring - The XPath string to use.
    component - A string representing which TDL component is being
                looked for (used in error reporting).
    optional  - A boolean describing whether this XML node is allowed to be
                absent or not.  If optional is True and the node is absent,
                None is returned.  If optional is False and the node is
                absent, an exception is raised.  (default: False)

    Returns the content of the XML node if found, None if the node is not
    found and optional is True.
    """
    res = doc.xpath(xmlstring)
    if len(res) == 1:
        return res[0].text
    elif not res:
        if optional:
            return None
        else:
            raise oz.OzException.OzException("Failed to find %s in TDL" % (component))
    else:
        raise oz.OzException.OzException("Expected 0 or 1 %s in TDL, saw %d" % (component, len(res)))


def data_from_type(name, contenttype, content):
    '''
    A function to get data out of some content, possibly decoding it depending
    on the content type.  This function understands three types of content:
    raw (where no decoding is necessary), base64 (where the data needs to be
    base64 decoded), and url (where the data needs to be downloaded).  Because
    the data might be large, all data is sent to file handle, which is returned
    from the function.
    '''

    out = tempfile.NamedTemporaryFile()
    if contenttype == 'raw':
        out.write(content.encode('utf-8'))
    elif contenttype == 'base64':
        if sys.version_info.major == 2:
            out.write(base64.decodestring(content))
        else:
            out.write(base64.decodebytes(content.encode('utf-8')))
    elif contenttype == 'url':
        url = urlparse.urlparse(content)
        if url.scheme == "file":
            with open(url.netloc + url.path) as f:
                outstr = "".join(f.readlines())
                out.write(outstr.encode('utf-8'))
        else:
            oz.ozutil.http_download_file(content, out.fileno(), False, None)
    else:
        raise oz.OzException.OzException("Type for %s must be 'raw', 'url' or 'base64'" % (name))

    # make sure the data is flushed to disk for uses of the file through
    # the name
    out.flush()
    out.seek(0)

    return out


class Repository(object):
    """
    Class that represents a single repository to be used for installing
    packages.  Objects of this type contain 3 pieces of information:

    name       - The name of this repository.
    url        - The URL of this repository.
    signed     - Whether this repository is signed
    persisted  - Whether this repository should remain in the final image
    sslverify  - Whether SSL certificates should be verified
    """
    def __init__(self, name, url, signed, persisted, sslverify):
        self.name = name
        self.url = url
        self.signed = signed
        self.persisted = persisted
        self.sslverify = sslverify


class Package(object):
    """
    Class that represents a single package to be installed.
    Objects of this type contain 4 pieces of information:

    name     - The name of the package.
    repo     - The repository that this package comes from (optional).
    filename - The filename that contains this package (optional).
    args     - Arguments necessary to install this package (optional).
    """
    def __init__(self, name, repo, filename, args):
        self.name = name
        self.repo = repo
        self.filename = filename
        self.args = args


class ISOExtra(object):
    """
    Class that represents an extra element to add to an installation ISO.
    Objects of this type contain 3 pieces of information:

    element_type - "file" or "directory"
    source       - A source URL for the element.
    destination  - A relative destination for the element.
    """
    def __init__(self, element_type, source, destination):
        self.element_type = element_type
        self.source = source
        self.destination = destination


class TDL(object):
    """
    Class that represents a parsed piece of TDL XML.  Objects of this kind
    contain 10 pieces of information:

    name         - The name assigned to this TDL.
    distro       - The type of operating system this TDL represents.
    update       - The version of the operating system this TDL represents.
    arch         - The architecture of the operating system this TDL
                   represents. Currently this must be one of "i386" or
                   "x86_64".
    key          - The installation key necessary to install this operating
                   system (optional).
    description  - A free-form description of this TDL (optional).
    installtype  - The method to be used to install this operating system.
                   Currently this must be one of "url" or "iso".
    packages     - A list of Package objects describing the packages to be
                   installed on the operating system.  This list may be
                   empty.
    repositories - A dictionary of Repository objects describing the
                   repositories to be searched to find packages.  The
                   dictionary is indexed by repository name.  This
                   dictionary may be empty.
    files        - A dictionary of file contents to be added to the
                   operating system.  The dictionary is indexed by filename.
    commands     - A dictionary of commands to run inside the guest VM.  The
                   dictionary is indexed by commands.  This dictionary may
                   be empty.
    """
    def __init__(self, xmlstring, rootpw_required=False):
        # open the XML document
        tree = lxml.etree.parse(StringIO(xmlstring))
        tree.xinclude()
        self.doc = tree.getroot()

        # then validate the schema
        relaxng = lxml.etree.RelaxNG(file=os.path.join(os.path.dirname(__file__),
                                                       'tdl.rng'))
        valid = relaxng.validate(self.doc)
        if not valid:
            errstr = "\nXML schema validation failed:\n"
            for error in relaxng.error_log:
                errstr += "\tline %s: %s\n" % (error.line, error.message)
            raise oz.OzException.OzException(errstr)

        template = self.doc.xpath('/template')
        if len(template) != 1:
            raise oz.OzException.OzException("Expected 1 template section in TDL, saw %d" % (len(template)))
        self.version = template[0].get('version')

        if self.version:
            self._validate_tdl_version()

        self.name = _xml_get_value(self.doc, '/template/name', 'template name')

        self.distro = _xml_get_value(self.doc, '/template/os/name', 'OS name')

        self.update = _xml_get_value(self.doc, '/template/os/version',
                                     'OS version')

        self.arch = _xml_get_value(self.doc, '/template/os/arch',
                                   'OS architecture')
        if self.arch not in ["i386", "x86_64", "ppc64", "ppc64le", "aarch64", "armv7l", "s390x"]:
            raise oz.OzException.OzException("Architecture must be one of 'i386, x86_64, ppc64, ppc64le, armv7l, aarch64, or s390x'")

        self.key = _xml_get_value(self.doc, '/template/os/key', 'OS key',
                                  optional=True)
        # key is not required, so it is not fatal if it is None

        self.description = _xml_get_value(self.doc, '/template/description',
                                          'description', optional=True)
        # description is not required, so it is not fatal if it is None

        install = self.doc.xpath('/template/os/install')
        if len(install) != 1:
            raise oz.OzException.OzException("Expected 1 OS install section in TDL, saw %d" % (len(install)))
        self.installtype = install[0].get('type')
        if self.installtype is None:
            raise oz.OzException.OzException("Failed to find OS install type in TDL")

        # we only support md5/sha1/sha256 sums for ISO install types.  However,
        # we make sure the instance variables are set to None for both types
        # so code lower down in the stack doesn't have to care about the ISO
        # vs. URL install type distinction, and can just check whether or not
        # these URLs are None
        self.iso_md5_url = None
        self.iso_sha1_url = None
        self.iso_sha256_url = None

        if self.installtype == "url":
            self.url = _xml_get_value(self.doc, '/template/os/install/url',
                                      'OS install URL')
        elif self.installtype == "iso":
            self.iso = _xml_get_value(self.doc, '/template/os/install/iso',
                                      'OS install ISO')
            self.iso_md5_url = _xml_get_value(self.doc,
                                              '/template/os/install/md5sum',
                                              'OS install ISO MD5SUM',
                                              optional=True)
            self.iso_sha1_url = _xml_get_value(self.doc,
                                               '/template/os/install/sha1sum',
                                               'OS install ISO SHA1SUM',
                                               optional=True)
            self.iso_sha256_url = _xml_get_value(self.doc,
                                                 '/template/os/install/sha256sum',
                                                 'OS install ISO SHA256SUM',
                                                 optional=True)
            # only one of md5, sha1, or sha256 can be specified; raise an error
            # if multiple are
            if (self.iso_md5_url and self.iso_sha1_url) or (self.iso_md5_url and self.iso_sha256_url) or (self.iso_sha1_url and self.iso_sha256_url):
                raise oz.OzException.OzException("Only one of <md5sum>, <sha1sum>, and <sha256sum> can be specified")
        else:
            raise oz.OzException.OzException("Unknown install type " + self.installtype + " in TDL")

        self.rootpw = _xml_get_value(self.doc, '/template/os/rootpw',
                                     "root/Administrator password",
                                     optional=not rootpw_required)

        self.packages = []
        self._add_packages(self.doc.xpath('/template/packages/package'))

        self.files = {}
        for afile in self.doc.xpath('/template/files/file'):
            name = afile.get('name')
            if name is None:
                raise oz.OzException.OzException("File without a name was given")
            contenttype = afile.get('type')
            if contenttype is None:
                contenttype = 'raw'

            content = afile.text
            if content:
                content = content.strip()
            else:
                content = ''
            self.files[name] = data_from_type(name, contenttype, content)

        self.isoextras = self._add_isoextras('/template/os/install/extras/directory',
                                             'directory')
        self.isoextras += self._add_isoextras('/template/os/install/extras/file',
                                              'file')

        self.repositories = {}
        self._add_repositories(self.doc.xpath('/template/repositories/repository'))

        self.commands = self._parse_commands('/template/commands')
        self.precommands = self._parse_commands('/template/precommands')

        self.disksize = self._parse_disksize()

        self.icicle_extra_cmd = _xml_get_value(self.doc,
                                               '/template/os/icicle/extra_command',
                                               "extra icicle command",
                                               optional=True)

        self.kernel_param = _xml_get_value(self.doc,
                                           '/template/os/kernelparam',
                                           'custom kernel parameter',
                                           optional=True)

    def _parse_disksize(self):
        """
        Internal method to parse the disk size out of the TDL.
        """
        size = _xml_get_value(self.doc, '/template/disk/size', 'disk size',
                              optional=True)
        if size is None:
            # if it wasn't specified, return None; the Guest object will assign
            # a sensible default
            return None

        match = re.match(r'([0-9]*) *([GT]?)$', size)
        if not match or len(match.groups()) != 2:
            raise oz.OzException.OzException("Invalid disk size; it must be specified as a size in gigabytes, optionally suffixed with 'G' or 'T'")

        number = match.group(1)
        suffix = match.group(2)

        if not suffix or suffix == 'G':
            # for backwards compatibility, we assume G when there is no suffix
            size = number
        elif suffix == 'T':
            size = str(int(number) * 1024)

        return size

    def _parse_commands(self, xpath):
        """
        Internal method to parse the commands XML and put it into order.  This
        order can either be via parse order (implicit) or by using the
        'position' attribute in the commands XML (explicit).  Note that the two
        cannot be mixed; if position is specified on one node, it must be
        specified on all of them.  Conversely, if position is *not* specified
        on one node, it must *not* be specified on any of them.  Also note that
        if explicit ordering is used, it must be strictly sequential, starting
        at 1, with no duplicate numbers.
        """
        tmp = []
        saw_position = False
        for command in self.doc.xpath(xpath + "/command"):
            name = command.get('name')
            if name is None:
                raise oz.OzException.OzException("Command without a name was given")
            contenttype = command.get('type')
            if contenttype is None:
                contenttype = 'raw'

            content = ""
            if command.text:
                content = command.text.strip()
            if not content:
                raise oz.OzException.OzException("Empty commands are not allowed")

            # since XML doesn't *guarantee* an order, the correct way to
            # specify a particular order of commands is to use the "position"
            # attribute.  For backwards compatibility, if the order is not
            # specified, we just use the parse order.  That being said, we do
            # not allow a mix of position attributes and implicit order.  If
            # you use the position attribute on one command, you must use it
            # on all commands, and vice-versa.

            position = command.get('position')
            if position is not None:
                saw_position = True
                position = int(position)

            fp = data_from_type(name, contenttype, content)
            tmp.append((position, fp))

        commands = []
        if not saw_position:
            for pos, fp in tmp:
                commands.append(fp)
        else:
            tmp.sort(key=lambda x: x[0])
            order = 1
            for pos, fp in tmp:
                if pos is None:
                    raise oz.OzException.OzException("All command elements must have a position (explicit order), or none of them may (implicit order)")
                elif pos != order:
                    # this handles both the case where there are duplicates and
                    # the case where there is a missing number
                    raise oz.OzException.OzException("Cannot have duplicate or sparse command position order!")
                order += 1
                commands.append(fp)

        return commands

    def merge_packages(self, packages):
        """
        Method to merge additional packages into an existing TDL.  The
        packages argument should be a properly structured <packages/> string
        as explained in the TDL documentation.  If a package with the same
        name is in the existing TDL and in packages, the value in packages
        overrides.
        """
        packsdoc = lxml.etree.fromstring(packages)
        packslist = packsdoc.xpath('/packages/package')
        self._add_packages(packslist, True)

    def _add_packages(self, packslist, remove_duplicates=False):
        """
        Internal method to add the list of lxml.etree nodes from packslist into
        the self.packages array.  If remove_duplicates is False (the default),
        then a package that is listed both in packslist and in the initial
        TDL is listed twice.  If it is set to True, then a package that is
        listed both in packslist and the initial TDL is listed only once,
        from the packslist.
        """
        for package in packslist:
            # package name
            name = package.get('name')

            if name is None:
                raise oz.OzException.OzException("Package without a name was given")

            # repository that the package lives in (optional)
            repo = _xml_get_value(package, 'repository',
                                  "package repository section", optional=True)

            # filename of the package (optional)
            filename = _xml_get_value(package, 'file', "package filename",
                                      optional=True)

            # arguments to install package (optional)
            args = _xml_get_value(package, 'arguments', "package arguments",
                                  optional=True)

            if remove_duplicates:
                # delete any existing packages with this name
                for p in [package for package in self.packages if package.name == name]:
                    self.packages.remove(p)

            # now add in our new package def
            self.packages.append(Package(name, repo, filename, args))

    def merge_repositories(self, repos):
        """
        Method to merge additional repositories into an existing TDL.  The
        repos argument should be a properly structured <repositories/>
        string as explained in the TDL documentation.  If a repository with
        the same name is in the existing TDL and in repos, the value in
        repos overrides.
        """
        reposdoc = lxml.etree.fromstring(repos)
        reposlist = reposdoc.xpath('/repositories/repository')
        self._add_repositories(reposlist)

    def _add_repositories(self, reposlist):
        """
        Internal method to add the list of lxml.etree nodes from reposlist into
        the self.repositories dictionary.
        """
        def _get_optional_repo_bool(repo, name, default):
            """
            Internal method to get an option boolean from a repo XML section.
            """
            xmlstr = _xml_get_value(repo, name, name, optional=True)
            if xmlstr is None:
                xmlstr = default

            val = oz.ozutil.string_to_bool(xmlstr)
            if val is None:
                raise oz.OzException.OzException("Repository %s property must be 'true', 'yes', 'false', or 'no'" % (name))
            return val

        for repo in reposlist:
            name = repo.get('name')
            if name is None:
                raise oz.OzException.OzException("Repository without a name was given")
            url = _xml_get_value(repo, 'url', 'repository url')

            if urlparse.urlparse(url)[1] in ["localhost", "127.0.0.1",
                                             "localhost.localdomain"]:
                raise oz.OzException.OzException("Repositories cannot be localhost, since they must be reachable from the guest operating system")

            signed = _get_optional_repo_bool(repo, 'signed', 'no')

            persist = _get_optional_repo_bool(repo, 'persisted', 'yes')

            sslverify = _get_optional_repo_bool(repo, 'sslverify', 'no')

            # no need to delete - if the name matches we just overwrite here
            self.repositories[name] = Repository(name, url, signed, persist,
                                                 sslverify)

    def _add_isoextras(self, extraspath, element_type):
        """
        Internal method to add the list of extra ISO elements from the specified
        XML path into the self.isoextras list.
        """
        isoextras = []
        extraslist = self.doc.xpath(extraspath)
        if self.installtype != 'iso' and extraslist:
            raise oz.OzException.OzException("Extra ISO data can only be used with iso install type")

        for extra in extraslist:
            source = extra.get('source')
            if source is None:
                raise oz.OzException.OzException("Extra ISO element without a source was given")
            destination = extra.get('destination')
            if destination is None:
                raise oz.OzException.OzException("Extra ISO element without a destination was given")

            isoextras.append(ISOExtra(element_type, source, destination))

        return isoextras

    # I declare we will use a 2 element version string with a dot
    # This allows simple comparison by conversion to float
    schema_version = "1.0"

    def _validate_tdl_version(self):
        """
        Internal method to validate that we support the TDL version.
        """
        if float(self.version) > float(self.schema_version):
            raise oz.OzException.OzException("TDL version (%s) is higher than our known version (%s)" % (self.version, self.schema_version))
